"""Module to handle ab initio electrostatic potentials from the DFT code GPAW."""
import warnings
from typing import Sequence, Tuple, Union

import numpy as np
from ase import units
from ase.build.tools import cut
from scipy.interpolate import interp1d, interpn

from abtem.base_classes import Grid
from abtem.device import get_device_function
from abtem.potentials import AbstractPotentialBuilder, PotentialArray, _disc_meshgrid, pad_atoms, \
    PotentialIntegrator
from abtem.structures import orthogonalize_cell, rotate_atoms_to_plane, plane_to_axes
from abtem.utils import subdivide_into_batches

try:
    from gpaw.atom.shapefunc import shape_functions
    from gpaw import GPAW
except ImportError:
    warnings.warn('This functionality of abTEM requires GPAW, see https://wiki.fysik.dtu.dk/gpaw/.')


def interpolate_rectangle(array: np.ndarray,
                          cell: np.ndarray,
                          extent: Sequence[float],
                          gpts: Sequence[int],
                          origin: Sequence[float] = None):
    """
    Interpolation to rectangle function

    A function to interpolate an array to a given rectangle, here used to convert electrostatic potentials
    from non-orthogonal cells to rectangular ones for use in abTEM multislice simulations.

    :param array: Electrostatic potential array to be interpolated.
    :param cell: ASE atoms simulation cell.
    :param extent: Extent of the rectangle [Å].
    :param gpts: Number of interpolation grid points.
    :param origin: Origin of the rectangle. Default is (0,0).
    """

    if origin is None:
        origin = (0., 0.)

    origin = np.array(origin)
    extent = np.array(extent)

    P = np.array(cell)
    P_inv = np.linalg.inv(P)
    origin_t = np.dot(origin, P_inv)
    origin_t = origin_t % 1.0

    lower = np.dot(origin_t, P)
    upper = lower + extent

    padded_array = np.zeros((array.shape[0] + 1, array.shape[1] + 1))
    padded_array[:-1, :-1] = array
    padded_array[-1, :] = padded_array[0, :]
    padded_array[:, -1] = padded_array[:, 0]

    x = np.linspace(0, 1, padded_array.shape[0], endpoint=True)
    y = np.linspace(0, 1, padded_array.shape[1], endpoint=True)

    x_ = np.linspace(lower[0], upper[0], gpts[0], endpoint=False)
    y_ = np.linspace(lower[1], upper[1], gpts[1], endpoint=False)
    x_, y_ = np.meshgrid(x_, y_, indexing='ij')

    p = np.array([x_.ravel(), y_.ravel()]).T
    p = np.dot(p, P_inv) % 1.0

    interpolated = interpn((x, y), padded_array, p, method='splinef2d')
    return interpolated.reshape((gpts[0], gpts[1]))


def interpolate_cube(array, old_cell, new_cell, new_gpts, origin=None):
    from scipy.interpolate import RegularGridInterpolator

    if origin is None:
        origin = (0., 0., 0.)

    padded_array = np.zeros((array.shape[0] + 1, array.shape[1] + 1, array.shape[2] + 1))
    padded_array[:-1, :-1, :-1] = array
    padded_array[-1] = padded_array[0]
    padded_array[:, -1] = padded_array[:, 0]
    padded_array[:, :, -1] = padded_array[:, :, 0]

    x = np.linspace(0, 1, padded_array.shape[0], endpoint=True)
    y = np.linspace(0, 1, padded_array.shape[1], endpoint=True)
    z = np.linspace(0, 1, padded_array.shape[2], endpoint=True)

    interpolator = RegularGridInterpolator((x, y, z), padded_array)

    x = np.linspace(origin[0], origin[0] + new_cell[0], new_gpts[0], endpoint=False)
    y = np.linspace(origin[1], origin[1] + new_cell[1], new_gpts[1], endpoint=False)
    z = np.linspace(origin[2], origin[2] + new_cell[2], new_gpts[2], endpoint=False)

    x, y, z = np.meshgrid(x, y, z, indexing='xy')

    points = np.array([x.ravel(), y.ravel(), z.ravel()]).T

    P = np.array(old_cell)
    P_inv = np.linalg.inv(P)

    scaled_points = np.dot(points, P_inv) % 1.0
    interpolated = interpolator(scaled_points)

    return interpolated.reshape(new_gpts)


def get_paw_corrections(atom_index: int, calculator, rcgauss: float = 0.005) -> Tuple[np.ndarray, np.ndarray]:
    """
    PAW corrections function

    Function to calculate the projector-augmented wave corrections to the electrostatic potential, needed to
    calculate the all-electron potential from a converged calculation. This is implemented independently in
    abTEM to enable dealing with non-orthogonal cells, and to allow working with slices of large potentials.

    Parameters
    ----------
    atom_index: int
        Index of the atom for which the corrections are calculated.
    calculator: GPAW object
        Converged GPAW calculation.
    rcgauss: float
        Radius of the Gaussian smearing of the nuclear potentials [Å]. Default value is 0.005 Å.

    Returns
    -------
    two 1d arrays
        The evaluation points and values of the core contribution to the electronstatic potential.
    """

    dens = calculator.density
    dens.D_asp.redistribute(dens.atom_partition.as_serial())
    dens.Q_aL.redistribute(dens.atom_partition.as_serial())

    D_sp = dens.D_asp[atom_index]

    setup = dens.setups[atom_index]
    c = setup.xc_correction
    rgd = c.rgd
    ghat_g = shape_functions(rgd, **setup.data.shape_function, lmax=0)[0]
    Z_g = shape_functions(rgd, 'gauss', rcgauss, lmax=0)[0] * setup.Z
    D_q = np.dot(D_sp.sum(0), c.B_pqL[:, :, 0])
    dn_g = np.dot(D_q, (c.n_qg - c.nt_qg)) * np.sqrt(4 * np.pi)
    dn_g += 4 * np.pi * (c.nc_g - c.nct_g)
    dn_g -= Z_g
    dn_g -= dens.Q_aL[atom_index][0] * ghat_g * np.sqrt(4 * np.pi)
    dv_g = rgd.poisson(dn_g) / np.sqrt(4 * np.pi)
    dv_g[1:] /= rgd.r_g[1:]
    dv_g[0] = dv_g[1]
    dv_g[-1] = 0.0

    return rgd.r_g, dv_g


class GPAWPotential(AbstractPotentialBuilder):
    """
    GPAW DFT potential object

    The GPAW potential object is used to calculate electrostatic potential of a converged GPAW calculator object.

    Parameters
    ----------
    calculator: GPAW object
        A converged GPAW calculator.
    origin: two float, optional
        xy-origin of the electrostatic potential relative to the xy-origin of the Atoms object [Å].
    gpts: one or two int
        Number of grid points describing each slice of the potential.
    sampling: one or two float
        Lateral sampling of the potential [1 / Å].
    slice_thickness: float
        Thickness of the potential slices in Å for calculating the number of slices used by the multislice algorithm.
        Default is 0.5 Å.
    core_size: float
        The standard deviation of the Gaussian function representing the atomic core.
    """

    def __init__(self,
                 calculator,
                 gpts: Union[int, Sequence[int]] = None,
                 sampling: Union[float, Sequence[float]] = None,
                 origin: Union[float, Sequence[float]] = None,
                 orthogonal_cell: Sequence[float] = None,
                 periodic_z: bool = True,
                 slice_thickness=.5,
                 core_size=.005,
                 plane='xy',
                 storage='cpu',
                 precalculate=True):

        self._calculator = calculator
        self._core_size = core_size
        self._plane = plane

        if orthogonal_cell is None:
            atoms = rotate_atoms_to_plane(calculator.atoms, plane)
            thickness = atoms.cell[2, 2]
            nz = calculator.hamiltonian.finegd.N_c[plane_to_axes(plane)[-1]]
            extent = np.diag(orthogonalize_cell(atoms.copy()).cell)[:2]
        else:
            if plane != 'xy':
                raise NotImplementedError()

            thickness = orthogonal_cell[2]
            nz = calculator.hamiltonian.finegd.N_c / np.linalg.norm(calculator.atoms.cell, axis=0) * orthogonal_cell[2]
            nz = int(np.ceil(np.max(nz)))
            extent = orthogonal_cell[:2]

        num_slices = int(np.ceil(nz / np.floor(slice_thickness / (thickness / nz))))
        self._orthogonal_cell = orthogonal_cell
        self._voxel_height = thickness / nz
        self._slice_vertical_voxels = subdivide_into_batches(nz, num_slices)
        self._origin = (0., 0., 0.)
        self._periodic_z = periodic_z

        self._grid = Grid(extent=extent, gpts=gpts, sampling=sampling, lock_extent=True)

        super().__init__(precalculate=precalculate, storage=storage)

    @property
    def calculator(self):
        return self._calculator

    @property
    def num_frozen_phonon_configs(self):
        return 1

    def generate_frozen_phonon_potentials(self, pbar=False):
        for i in range(self.num_frozen_phonon_configs):
            if self._precalculate:
                yield self.build(pbar=pbar)
            else:
                yield self

    @property
    def core_size(self):
        return self._core_size

    @property
    def origin(self):
        return self._origin

    @property
    def num_slices(self):
        return len(self._slice_vertical_voxels)

    def get_slice_thickness(self, i):
        return self._slice_vertical_voxels[i] * self._voxel_height

    def generate_slices(self, first_slice=0, last_slice=None, max_batch=1):
        interpolate_radial_functions = get_device_function(np, 'interpolate_radial_functions')

        if last_slice is None:
            last_slice = len(self)

        if self._plane != 'xy':
            atoms = rotate_atoms_to_plane(self._calculator.atoms.copy(), self._plane)
        else:
            atoms = self._calculator.atoms.copy()

        old_cell = atoms.cell

        atoms.set_tags(range(len(atoms)))

        if self._orthogonal_cell is None:
            atoms = orthogonalize_cell(atoms)
        else:
            scaled = atoms.cell.scaled_positions(np.diag(self._orthogonal_cell))
            atoms = cut(atoms, a=scaled[0], b=scaled[1], c=scaled[2])

        valence = self._calculator.get_electrostatic_potential()
        new_gpts = self.gpts + (sum(self._slice_vertical_voxels),)

        axes = plane_to_axes(self._plane)
        if self._plane != 'xy':
            array = np.moveaxis(valence, axes[:2], (0, 1))
        else:
            array = valence

        from scipy.interpolate import RegularGridInterpolator

        origin = (0., 0., 0.)

        padded_array = np.zeros((array.shape[0] + 1, array.shape[1] + 1, array.shape[2] + 1))
        padded_array[:-1, :-1, :-1] = array
        padded_array[-1] = padded_array[0]
        padded_array[:, -1] = padded_array[:, 0]
        padded_array[:, :, -1] = padded_array[:, :, 0]

        x = np.linspace(0, 1, padded_array.shape[0], endpoint=True)
        y = np.linspace(0, 1, padded_array.shape[1], endpoint=True)
        z = np.linspace(0, 1, padded_array.shape[2], endpoint=True)

        interpolator = RegularGridInterpolator((x, y, z), padded_array)

        new_cell = np.diag(atoms.cell)
        x = np.linspace(origin[0], origin[0] + new_cell[0], new_gpts[0], endpoint=False)
        y = np.linspace(origin[1], origin[1] + new_cell[1], new_gpts[1], endpoint=False)
        z = np.linspace(origin[2], origin[2] + new_cell[2], new_gpts[2], endpoint=False)

        P = np.array(old_cell)
        P_inv = np.linalg.inv(P)

        cutoffs = {}
        for number in np.unique(atoms.numbers):
            indices = np.where(atoms.numbers == number)[0]
            r = self._calculator.density.setups[indices[0]].xc_correction.rgd.r_g[1:] * units.Bohr
            cutoffs[number] = r[-1]

        if self._periodic_z:
            atoms = pad_atoms(atoms, margin=max(cutoffs.values()), directions='z', in_place=True)

        indices_by_number = {number: np.where(atoms.numbers == number)[0] for number in np.unique(atoms.numbers)}

        na = sum(self._slice_vertical_voxels[:first_slice])
        a = na * self._voxel_height
        for i in range(first_slice, last_slice):
            nb = na + self._slice_vertical_voxels[i]
            b = a + self._slice_vertical_voxels[i] * self._voxel_height

            X, Y, Z = np.meshgrid(x, y, z[na:nb], indexing='ij')

            points = np.array([X.ravel(), Y.ravel(), Z.ravel()]).T

            scaled_points = np.dot(points, P_inv) % 1.0

            projected_valence = interpolator(scaled_points).reshape(self.gpts + (nb - na,)).sum(
                axis=-1) * self._voxel_height

            array = np.zeros((1,) + self.gpts, dtype=np.float32)
            for number, indices in indices_by_number.items():
                slice_atoms = atoms[indices]

                if len(slice_atoms) == 0:
                    continue

                cutoff = cutoffs[number]
                margin = np.int(np.ceil(cutoff / np.min(self.sampling)))
                rows, cols = _disc_meshgrid(margin)
                disc_indices = np.hstack((rows[:, None], cols[:, None]))

                slice_atoms = slice_atoms[(slice_atoms.positions[:, 2] > a - cutoff) *
                                          (slice_atoms.positions[:, 2] < b + cutoff)]

                slice_atoms = pad_atoms(slice_atoms, margin=cutoff, directions='xy', )

                R = np.geomspace(np.min(self.sampling) / 2, cutoff, int(np.ceil(cutoff / np.min(self.sampling))) * 10)

                vr = np.zeros((len(slice_atoms), len(R)), np.float32)
                dvdr = np.zeros((len(slice_atoms), len(R)), np.float32)
                # TODO : improve speed of this
                for j, atom in enumerate(slice_atoms):
                    r, v = get_paw_corrections(atom.tag, self._calculator, self._core_size)

                    f = interp1d(r * units.Bohr, v, fill_value=(v[0], 0), bounds_error=False, kind='linear')

                    integrator = PotentialIntegrator(f, R, self.get_slice_thickness(i), tolerance=1e-6)

                    vr[j], dvdr[j] = integrator.integrate(np.array([atom.z]), a, b)

                sampling = np.asarray(self.sampling, dtype=np.float32)
                run_length_enconding = np.zeros((2,), dtype=np.int32)
                run_length_enconding[1] = len(slice_atoms)

                interpolate_radial_functions(array,
                                             run_length_enconding,
                                             disc_indices,
                                             slice_atoms.positions,
                                             vr,
                                             R,
                                             dvdr,
                                             sampling)

            array = -(projected_valence + array / np.sqrt(4 * np.pi) * units.Ha)

            yield i, i + 1, PotentialArray(array, np.array([self.get_slice_thickness(i)]), extent=self.extent)

            a = b
            na = nb

    def __copy__(self):
        slice_thickness = self.calculator.atoms.cell[2, 2] / self.num_slices
        return self.__class__(self.calculator,
                              gpts=self.gpts,
                              sampling=self.sampling,
                              # origin=self.origin,
                              slice_thickness=slice_thickness,
                              core_size=self.core_size,
                              storage=self._storage)



import numpy as np
from ase import Atoms

from abtem.structures import standardize_cell


def angle_between(v1, v2):
    v1 = v1 / np.linalg.norm(v1)
    v2 = v2 / np.linalg.norm(v2)
    return np.math.atan2(np.cross(v1, v2), np.dot(v1, v2))


def graphene_bilayer(n, a=2.46, d=3):
    m = n - 1
    gcd = np.gcd(n, m)
    n = int(n / gcd)
    m = int(m / gcd)

    positions = np.array([[0, 0], [0, a / np.sqrt(3)]])
    cell = np.array([[a, 0], [0.5 * a, 1.5 * a / np.sqrt(3)]])

    for i in range(2 * n - 3):
        if (i == 0) or (i % 2 == 1):
            positions = np.append(positions, (positions[i * 2] + cell[1]).reshape(1, -1), axis=0)
            positions = np.append(positions, (positions[i * 2 + 1] + cell[1]).reshape(1, -1), axis=0)
        if (i == 0) or (i % 2 == 0):
            positions = np.append(positions, (positions[i * 2] + cell[0] - cell[1]).reshape(1, -1), axis=0)
            positions = np.append(positions, (positions[i * 2 + 1] + cell[0] - cell[1]).reshape(1, -1), axis=0)

    for i in range(2 * (n - 1) * (2 * n - 1)):
        positions = np.append(positions, (positions[i] + cell[0]).reshape(1, -1), axis=0)

    last = 2 * (n - 1) * (2 * n - 1)
    side = 2 * (2 * n - 1)
    for i in range(n - 1):
        for j in range(last, last + side - 4):
            positions = np.append(positions, (positions[j] + cell[0]).reshape(1, -1), axis=0)
        last = last + side
        side = side - 4

    v1 = m * cell[0] + n * cell[1]
    v2 = n * cell[0] + m * cell[1]
    theta = angle_between(v2, v1)

    R = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])

    v2 = np.dot(R, v2)
    v3 = np.dot(R, -m * cell[0] + (n + m) * cell[1])

    super_cell = np.array([np.append(v2, 0), np.append(v3, 0), [0, 0, 2 * d]])

    rotated_positions = positions + np.array([-0.5 * a, 0.5 * a / np.sqrt(3)])
    rotated_positions = np.dot(R, rotated_positions.T).T

    positions = np.hstack((positions, np.ones(len(positions)).reshape(-1, 1)))
    rotated_positions = np.hstack((rotated_positions, d * np.ones(len(rotated_positions)).reshape(-1, 1)))

    atoms = Atoms('C' * len(positions), positions=positions, cell=super_cell, pbc=True)
    rotated_atoms = Atoms('C' * len(rotated_positions), positions=rotated_positions, cell=super_cell, pbc=True)

    atoms += rotated_atoms
    atoms.wrap()
    standardize_cell(atoms)
    return atoms



import ipywidgets as widgets
import numpy as np
from bqplot import Figure, LinearScale, Axis
from traitlets import HasTraits, observe, Dict, Instance, link, Bool, List, default, Float
from traitlets import Unicode


class Canvas(widgets.HBox):
    artists = Dict()
    tools = Dict()
    tool = Unicode('None')
    title = Unicode()
    x_label = Unicode()
    y_label = Unicode()
    x_limits = List()
    y_limits = List()
    lock_scale = Bool(True, allow_none=True)
    _tool_artists = List()

    def __init__(self,
                 x_scale=None,
                 y_scale=None,
                 x_axis=None,
                 y_axis=None,
                 height=450.,
                 width=450.,
                 fig_margin=None,
                 **kwargs):

        x_scale = x_scale or LinearScale(allow_padding=False)
        y_scale = y_scale or LinearScale(allow_padding=False)

        scales = {'x': x_scale, 'y': y_scale}

        x_axis = x_axis or Axis(scale=scales['x'])
        y_axis = y_axis or Axis(scale=scales['y'], orientation='vertical')

        fig_margin = fig_margin or {'top': 0, 'bottom': 50, 'left': 50, 'right': 0}

        margin_width = width - fig_margin['left'] - fig_margin['right']
        margin_height = width - fig_margin['bottom'] - fig_margin['top']

        min_aspect_ratio = width / height
        max_aspect_ratio = width / height

        figure = Figure(scales=scales,
                        axes=[x_axis, y_axis],
                        min_aspect_ratio=min_aspect_ratio,
                        max_aspect_ratio=max_aspect_ratio,
                        fig_margin=fig_margin)

        figure.layout.height = f'{height}px'
        figure.layout.width = f'{width}px'

        whitespace = widgets.HBox([])
        whitespace.layout.width = f'{figure.fig_margin["left"]}px'

        title = widgets.HTML(value=f"<p style='font-size:16px;text-align:center'> {self.title} </p>")
        title.layout.width = f'{float(figure.layout.width[:-2]) - figure.fig_margin["left"]}px'
        canvas_widgets = widgets.VBox([widgets.HBox([whitespace, title]), figure])
        self._figure = figure

        super().__init__(children=[canvas_widgets], **kwargs)
        link((self, 'x_label'), (x_axis, 'label'))
        link((self, 'y_label'), (y_axis, 'label'))

        # link((self, 'title'), (figure, 'title'))

    @property
    def figure(self):
        return self._figure

    @property
    def figure_width(self):
        return int(float(self.figure.layout.width[:-2]))

    @property
    def figure_height(self):
        return int(float(self.figure.layout.width[:-2]))

    def pixel_to_domain(self, x, y):
        pixel_margin_width = self.figure.fig_margin['left'] + self.figure.fig_margin['right']
        pixel_margin_height = self.figure.fig_margin['top'] + self.figure.fig_margin['bottom']

        pixel_width = self.figure_width - pixel_margin_width
        pixel_height = self.figure_height - pixel_margin_height

        domain_width = self.x_scale.max - self.x_scale.min
        domain_height = self.y_scale.max - self.y_scale.min

        x = domain_width / pixel_width * (x - self.figure.fig_margin['left']) + self.x_scale.min
        y = domain_height / pixel_height * (y - self.figure.fig_margin['top']) + self.y_scale.min
        y = - y + self.y_scale.max +self.y_scale.min
        return x, y

    @property
    def widget(self):
        whitespace = widgets.HBox([])
        whitespace.layout.width = f'{self.figure.fig_margin["left"]}px'

        title = widgets.HTML(value=f"<p style='font-size:16px;text-align:center'> {self.title} </p>")
        title.layout.width = f'{float(self.figure.layout.width[:-2]) - self.figure.fig_margin["left"]}px'
        return widgets.VBox([widgets.HBox([whitespace, title]), self.figure])

    @property
    def x_axis(self):
        return self.figure.axes[0]

    @property
    def y_axis(self):
        return self.figure.axes[1]

    @property
    def x_scale(self):
        return self.x_axis.scale

    @property
    def y_scale(self):
        return self.y_axis.scale

    @observe('artists', '_tool_artists')
    def _observe_artists(self, change):
        self._update_marks()

    def _update_marks(self):
        self.figure.marks = []

        for key, artist in self.artists.items():
            artist._add_to_canvas(self)

        for artist in self._tool_artists:
            artist._add_to_canvas(self)

        # adjust_x = False
        # adjust_y = False
        # if (self.x_scale.min is not None) or (self.x_scale.max is not None):
        #     adjust_x = True
        #
        # if (self.y_scale.min is not None) or (self.y_scale.max is not None):
        #     adjust_y = True

        # self.adjust_limits_to_artists(adjust_x=adjust_x, adjust_y=adjust_y)

    def _enforce_scale_lock(self, adjust_x=True, adjust_y=True):
        if not self.lock_scale:
            return

        if None in (self.x_scale.min, self.x_scale.min, self.y_scale.min, self.y_scale.max):
            return

        extent = max(self.x_scale.max - self.x_scale.min, self.y_scale.max - self.y_scale.min) / 2
        x_center = (self.x_scale.min + self.x_scale.max) / 2
        y_center = (self.y_scale.min + self.y_scale.max) / 2

        with self.x_scale.hold_trait_notifications(), self.y_scale.hold_trait_notifications():
            if adjust_x:
                self.x_scale.min = x_center - extent
                self.x_scale.max = x_center + extent

            if adjust_y:
                self.y_scale.min = y_center - extent
                self.y_scale.max = y_center + extent

    @observe('tool')
    def _observe_tool(self, change):
        if change['old'] != 'None':
            self.tools[change['old']].deactivate(self)

        if change['new'] == 'None':
            self.figure.interaction = None
        else:
            self.tools[change['new']].activate(self)

    @property
    def toolbar(self):
        tool_names = ['None'] + list(self.tools.keys())

        tool_selector = widgets.ToggleButtons(options=tool_names, value=self.tool)
        tool_selector.style.button_width = '60px'

        link((tool_selector, 'value'), (self, 'tool'))

        whitespace = widgets.HBox([])
        whitespace.layout.width = f'{self.figure.fig_margin["left"]}px'

        reset_button = widgets.Button(description='Reset', layout=widgets.Layout(width='80px'))
        reset_button.on_click(lambda _: self.adjust_limits_to_artists())

        return widgets.HBox([whitespace, widgets.HBox([tool_selector]), reset_button])

    @observe('x_limits')
    def _observe_x_limits(self, change):
        if change['new'] is None:
            self.x_scale.min = None
            self.x_scale.max = None
        else:
            self.x_scale.min = change['new'][0]
            self.x_scale.max = change['new'][1]

    @observe('y_limits')
    def _observe_y_limits(self, change):
        if change['new'] is None:
            self.y_scale.min = None
            self.y_scale.max = None
        else:
            self.y_scale.min = change['new'][0]
            self.y_scale.max = change['new'][1]

    @property
    def visibility_checkboxes(self):
        checkboxes = []
        for key, artist in self.artists.items():
            checkbox = widgets.Checkbox(value=True, description=key, indent=False, layout=widgets.Layout(width='90%'))
            link((checkbox, 'value'), (artist, 'visible'))
            checkboxes.append(checkbox)
        return widgets.VBox(checkboxes)

    def adjust_labels_to_artists(self):
        x_labels = [artist.x_label for artist in self.artists.values() if artist.x_label is not None]
        if len(x_labels) > 0:
            self.x_label = x_labels[0]

        y_labels = [artist.y_label for artist in self.artists.values() if artist.y_label is not None]
        if len(y_labels) > 0:
            self.y_label = y_labels[0]

    def adjust_limits_to_artists(self, adjust_x=True, adjust_y=True, *args):
        xmin = np.min([artist.limits[0][0] for artist in self.artists.values()])
        xmax = np.max([artist.limits[0][1] for artist in self.artists.values()])
        ymin = np.min([artist.limits[1][0] for artist in self.artists.values()])
        ymax = np.max([artist.limits[1][1] for artist in self.artists.values()])
        with self.x_scale.hold_trait_notifications(), self.y_scale.hold_trait_notifications():
            if adjust_x:
                self.x_limits = [float(xmin), float(xmax)]

            if adjust_y:
                self.y_limits = [float(ymin), float(ymax)]

            self._enforce_scale_lock()



"""Module to efficiently handle numerical integrals with singularities by a double exponential Tanh–Sinh quadrature."""
import numpy as np
import scipy.special
from typing import Callable, Sequence


def _error_estimate(eps: float,
                    value_estimates: Sequence[float],
                    left_summands: Sequence[float],
                    right_summands: Sequence[float]):
    # TODO: improve error estimate using the potential derivatives
    """
    Internal function to estimate the error made by the quadrature.

    Parameters
    ----------
    eps: float
        The error tolerance.
    value_estimates: list of float
        Sequence of estimated integral values
    left_summands: list of float
    right_summands: list of float

    Returns
    -------
    float
        The estimated error of the integration value (upper bound).
    """

    if len(value_estimates) < 3:
        error_estimate = 1

    elif value_estimates[0] == value_estimates[-1]:
        error_estimate = 0

    else:
        # e1 = abs(value_estimates[-1] - value_estimates[-2])
        # e2 = abs(value_estimates[-1] - value_estimates[-3])
        # e3 = eps * max(max(abs(left_summands)), max(abs(right_summands)))
        # e4 = max(abs(left_summands[-1]), abs(right_summands[-1]))
        # n = np.float((np.log(e1 + np.finfo(e1).eps) / (np.log(e2 + np.finfo(e2).eps))))

        e1 = abs(value_estimates[-1] - value_estimates[-2])
        e2 = abs(value_estimates[-1] - value_estimates[-3])
        e3 = eps * max(max(abs(left_summands)), max(abs(right_summands)))
        e4 = max(abs(left_summands[-1]), abs(right_summands[-1]))

        with np.errstate(over='ignore'):
            error_estimate = max(e1 ** (np.log(e1) / np.log(e2)), e1 ** 2, e3, e4)

        #
        #    error_estimate = max(e1 ** n, e1 ** 2, e3, e4)
        # error_estimate = max(e1, e4)

    return error_estimate


def _solve_expx_x_logx(tau, tol, max_steps=10):
    """Internal function to numerically calculate an auxiliary function."""
    x = np.log(2 / np.pi * np.log(np.pi / tau))

    def f0(x):
        return np.pi / 2 * np.exp(x) - x - np.log(x * np.pi / tau)

    def f1(x):
        return np.pi / 2 * np.exp(x) - 1 - 1 / x

    f0x = f0(x)
    success = False

    for _ in range(max_steps):
        x -= f0x / f1(x)
        f0x = f0(x)
        if abs(f0x) < tol:
            success = True
            break

    assert success
    return x


def integrate(f: Callable, a: float, b: float, eps: float, max_steps: int = 20, max_order=2000):
    """
    Integrate a function using the Tanh-Sinh quadrature method.

    Parameters
    ----------
    f: callable
        Function to integrate.
    a: float
        Lower integration limit.
    b: float
        Upper integration limit.
    eps: float
        Integration error tolerance.
    max_steps: int
        Maximum number of refinement steps allowed before an error is raised.

    Returns
    -------
    integral_value: float
        The estimated integration value.
    error_estimate: float
        The estimated error of the integration value (upper bound).
    step_size: float
        The quadrature step size used to obtain the final integral value.
    order: int
        The quadrature order used to obtain the final integral value.
    """

    def f_left(s):
        return f(a + s)

    def f_right(s):
        return f(b - s)

    interval = b - a

    def lambertw(x, k):
        out = scipy.special.lambertw(x, k)
        assert abs(out.imag) < 1.0e-15
        return scipy.special.lambertw(x, k).real

    h = _solve_expx_x_logx(eps ** 2, tol=1e-7)

    order = 0
    success = False
    for level in range(max_steps + 1):
        assert eps ** 2 * np.exp(np.pi / 2) < np.pi * h
        j = int(np.log(-2 / np.pi * lambertw(-(eps ** 2) / h / 2, -1)) / h)

        if level == 0:
            t = [0]
        else:
            t = h * np.arange(1, j + 1, 2)

        sinh_t = np.pi / 2 * np.sinh(t)
        cosh_t = np.pi / 2 * np.cosh(t)
        cosh_sinh_t = np.cosh(sinh_t)
        exp_sinh_t = np.exp(sinh_t)

        y0 = interval / 2 / exp_sinh_t / cosh_sinh_t

        weights = h * interval / 2 * cosh_t / cosh_sinh_t ** 2
        left_summands = f_left(y0) * weights
        right_summands = f_right(y0) * weights

        if level == 0:
            value_estimates = list(left_summands)
        else:
            value_estimates.append(value_estimates[-1] / 2 + np.sum(left_summands) + np.sum(right_summands))

        error_estimate = _error_estimate(eps, value_estimates, left_summands, right_summands)

        if order + j + 1 > max_order:
            success = True
            break

        if abs(error_estimate) < eps:
            success = True
            break

        order += j + 1
        h /= 2

    assert success
    return value_estimates[-1], error_estimate, h, order - 1


def tanh_sinh_nodes_and_weights(step_size: float, order: int):
    """
    Calculate the nodes and weights for the tanh-sinh quadrature with a given step size and order.

    Parameters
    ----------
    step_size: float
        The quadrature step size.
    order: int
        The quadrature order.

    Returns
    -------
    array, array
        Nodes and weights
    """

    xk = np.zeros(2 * order + 1, dtype=np.float)
    wk = np.zeros(2 * order + 1, dtype=np.float)
    for i, k in enumerate(range(-order, order + 1)):
        xk[i] = np.tanh(np.pi / 2 * np.sinh(k * step_size))
        numerator = step_size / 2 * np.pi * np.cosh(k * step_size)
        denominator = np.cosh(np.pi / 2 * np.sinh(k * step_size)) ** 2
        wk[i] = numerator / denominator
    return xk, wk



import numpy as np

from abtem import Probe, FresnelPropagator


def test_probe_tilt():
    tilted_probe = Probe(extent=5, gpts=128, energy=60e3, semiangle_cutoff=30, tilt=(200, 200))
    shifted_probe = Probe(extent=5, gpts=128, energy=60e3, semiangle_cutoff=30, tilt=(0, 0))

    tilted_wave = tilted_probe.build((0, 0))
    assert tilted_wave.tilt == tilted_probe.tilt

    propagator = FresnelPropagator()
    dz = 5
    tilted_wave = propagator.propagate(tilted_wave, dz)
    shifted_wave = shifted_probe.build([np.tan(tilted_wave.tilt[0] / 1e3) * dz, np.tan(tilted_wave.tilt[1] / 1e3) * dz])
    shifted_wave = propagator.propagate(shifted_wave, 5)
    assert np.allclose(shifted_wave.array, tilted_wave.array, atol=1e-5)


def test_probe_ctf():
    probe = Probe(extent=5, gpts=128, energy=60e3, semiangle_cutoff=30)
    defocused_probe = Probe(extent=5, gpts=128, energy=60e3, semiangle_cutoff=30, defocus=50)

    assert defocused_probe.ctf.defocus == 50

    assert not np.allclose(probe.build().array, defocused_probe.build().array)

    probe.ctf.defocus = 50
    assert np.allclose(probe.build().array, defocused_probe.build().array)



from pathlib import Path

import ipywidgets as widgets
from scipy.ndimage import gaussian_filter
from skimage.io import imread
from skimage.transform import downscale_local_mean
from traitlets import HasTraits, List, Unicode, Any, Int, observe, Union
from traittypes import Array

from abtem.visualize.interactive.utils import link


class GaussianFilterSlider(widgets.FloatSlider):
    image_in = Array()
    image_out = Array()

    def __init__(self, description='Gaussian filter', min=0., max=5, **kwargs):
        super().__init__(description=description, min=min, max=max, **kwargs)

    def update_image(self, *args):
        self.image_out = self(self.image_in)

    @observe('image_in')
    def _observe_image_in(self, *args):
        self.update_image()

    def __call__(self, image):
        return gaussian_filter(image, sigma=float(self.value))


class IntSliderWithButtons(widgets.HBox):
    value = Int(0)
    min = Int(0)
    max = Int(0)

    def __init__(self, **kwargs):
        self._slider = widgets.IntSlider(**kwargs)

        self._slider.layout.width = '300px'

        link((self._slider, 'value'), (self, 'value'))
        link((self._slider, 'min'), (self, 'min'))
        link((self._slider, 'max'), (self, 'max'))

        previous_button = widgets.Button(description='Previous')
        previous_button.layout.width = '100px'
        next_button = widgets.Button(description='Next')
        next_button.layout.width = '100px'

        def next_item(*args):
            self._slider.value += 1

        def previous_item(*args):
            self._slider.value -= 1

        next_button.on_click(next_item)
        previous_button.on_click(previous_item)

        super().__init__([self._slider, previous_button, next_button])


class ImageCollectionBrowser(HasTraits):
    paths = Union((List(), Unicode()))
    file_index = Int(0)
    path = Unicode()
    images = Array()
    image = Any()
    frame_index = Int(0)
    num_frames = Int(0)
    binning = Int(1)

    def __init__(self, **kwargs):
        self._path_text = widgets.HTML(description='File name')
        self._file_index_slider = IntSliderWithButtons(description='File index', min=0)
        self._frame_index_slider = IntSliderWithButtons(description='Frame index', min=0)
        link((self._file_index_slider, 'value'), (self, 'file_index'))
        link((self._frame_index_slider, 'value'), (self, 'frame_index'))
        super().__init__(**kwargs)

    @observe('frame_index')
    def _observe_frame_index(self, *args):
        image = self.images[self.frame_index]

        if self.binning != 1:
            image = downscale_local_mean(image, factors=(self.binning,) * 2)

        self.image = image

    @observe('path')
    def _observe_path(self, *args):
        self._path_text.value = str(Path(*Path(self.path).parts[-4:]))

        images = imread(self.path)

        # if images.shape[-1] in (3, 4):
        #    images = np.swapaxes(images, 0, 2)

        if len(images.shape) == 2:
            images = images[None]

        assert len(images.shape) == 3

        self.images = images

        self.num_frames = len(self.images)

        self._frame_index_slider.max = len(self.images) - 1

        if self.frame_index == 0:
            self._observe_frame_index()
        else:
            self.frame_index = 0

    @observe('file_index')
    def _observe_file_index(self, *args):
        self.path = str(self.paths[self.file_index])

    @observe('paths')
    def _observe_filename(self, change):

        self._file_index_slider.max = len(self.paths) - 1

        if self.file_index == 0:
            self._observe_file_index()
        else:
            self.file_index = 0

    @property
    def widgets(self):
        hbox = widgets.VBox([self._path_text, self._file_index_slider, self._frame_index_slider])
        return hbox



import numpy as np
import pytest

from abtem import GridScan, AnnularDetector, PixelatedDetector, Probe, Potential, PlaneWave, FlexibleAnnularDetector, \
    Measurement
from abtem.measure import center_of_mass
from abtem.noise import poisson_noise


def test_calibration_coordinates():
    for endpoint in (True, False):
        gridscan = GridScan(sampling=.7, start=(0, 0), end=(4, 4), endpoint=endpoint)
        calibration_coordinates = gridscan.calibrations[0].coordinates(gridscan.gpts[0])
        assert np.all(calibration_coordinates == gridscan.get_positions()[::gridscan.gpts[0], 0])


@pytest.fixture()
def stem_data(data_path, graphene_atoms):
    potential = Potential(atoms=graphene_atoms, sampling=.05)
    probe = Probe(energy=200e3, semiangle_cutoff=30)
    probe.grid.match(potential)

    scan = GridScan(start=[0, 0], end=[potential.extent[0], potential.extent[1]], gpts=(10, 9))

    pixelated_detector = PixelatedDetector()
    flexible_annular_detector = FlexibleAnnularDetector()
    annular_detector = AnnularDetector(50, 150)
    measurements = probe.scan(scan, [annular_detector, flexible_annular_detector, pixelated_detector], potential,
                              pbar=False)
    return {'annular': measurements[0], 'flexible': measurements[1], 'pixelated': measurements[2]}


@pytest.fixture()
def hrtem_image(data_path, graphene_atoms):
    potential = Potential(atoms=graphene_atoms, sampling=.05)
    planewave = PlaneWave(energy=200e3)
    planewave.grid.match(potential)

    exit_wave = planewave.multislice(potential, pbar=False)

    return exit_wave.intensity()


def test_poisson_noise(stem_data, hrtem_image):
    poisson_noise(stem_data['annular'], 1e3)
    poisson_noise(stem_data['pixelated'], 1e3)
    poisson_noise(hrtem_image, 1e3)


def test_integrate_flexible_annular_measurement(stem_data):
    integrated = stem_data['flexible'].integrate(50, 150)
    assert np.allclose(integrated.array, stem_data['annular'].array)


def test_add_measurements(stem_data):
    stem_data = stem_data['annular'].copy()
    stem_data = stem_data + stem_data
    stem_data += stem_data


def test_subtract_measurements(stem_data):
    stem_data = stem_data['annular'].copy()
    stem_data = stem_data - stem_data
    stem_data -= stem_data

@pytest.mark.hyperspy
@pytest.mark.parametrize("signal_type", [None, "diffraction"])
def test_to_hyperspy(stem_data, signal_type):
    for key in stem_data:
        sig = stem_data[key].to_hyperspy(signal_type=signal_type)
        assert isinstance(sig, hs.signals.BaseSignal)


@pytest.mark.hyperspy
def test_to_hyperspy_hrtem(hrtem_image):
    hrtem_image.to_hyperspy()


def test_read_write_measurement(tmp_path, stem_data, hrtem_image):
    d = tmp_path / 'sub'
    d.mkdir()
    path = d / 'measurement.hdf5'

    stem_data['annular'].write(path)
    Measurement.read(path)

    stem_data['flexible'].write(path)
    Measurement.read(path)

    stem_data['pixelated'].write(path)
    Measurement.read(path)

    #hrtem_image.write(path)
    #Measurement.read(path)


@pytest.mark.hyperspy
def test_write_hspy(tmp_path, stem_data, hrtem_image):
    d = tmp_path / 'sub'
    d.mkdir()
    path = d / 'measurement.hspy'

    stem_data['annular'].write(path, format="hspy")

    stem_data['flexible'].write(path, format="hspy")

    stem_data['pixelated'].write(path, format="hspy")

    #hrtem_image.write(path, format="hspy")


def test_indexing(stem_data):
    diffraction_pattern = stem_data['pixelated'][0, 0]
    assert len(diffraction_pattern.calibrations) == 2
    assert diffraction_pattern.calibrations[0].units == 'mrad'


def test_center_of_mass(stem_data):
    center_of_mass(stem_data['pixelated'])
    center_of_mass(stem_data['pixelated'], return_icom=True)



import numpy as np

from abtem.base_classes import Accelerator, HasAcceleratorMixin
from abtem.device import get_array_module


class BullseyeAperture(HasAcceleratorMixin):

    def __init__(self, outer_angle, energy=None, inner_angle=0., num_radials=0, cross=0., rotation=0.):
        self._outer_angle = outer_angle
        self._inner_angle = inner_angle
        self._num_radials = num_radials
        self._rotation = rotation
        self._cross = cross
        self._accelerator = Accelerator(energy=energy)

    def evaluate(self, alpha, phi):
        xp = get_array_module(alpha)

        aperture = xp.ones_like(alpha)

        alpha = alpha * 1000

        aperture[alpha < self._inner_angle] = 0.
        aperture[alpha > self._outer_angle] = 0.

        if self._num_radials > 0:
            edges = np.linspace(self._inner_angle, self._outer_angle, (self._num_radials + 1) * 2)

            start_edges = [edge for i, edge in enumerate(edges[:-1]) if i % 2]
            end_edges = [edge for i, edge in enumerate(edges[1:-1]) if i % 2]

            for start_edge, end_edge in zip(start_edges, end_edges):
                aperture[(alpha > start_edge) * (alpha < end_edge)] = 0.

        if self._cross > 0.:
            d = np.abs(np.sin(phi - self._rotation) * alpha)
            aperture[(d < self._cross / 2) * (alpha < self._outer_angle)] = 1.

            d = np.abs(np.sin(phi - self._rotation - np.pi / 2) * alpha)
            aperture[(d < self._cross / 2) * (alpha < self._outer_angle)] = 1.

        return aperture



from colorsys import hls_to_rgb

import numpy as np
from scipy.interpolate import interpn, interp1d
from abtem.visualize.colors import hsluv
from matplotlib.colors import ListedColormap


def format_label(calibration):
    if calibration is None:
        return ''

    label = ''
    if calibration.name:
        label += f'{calibration.name}'

    if calibration.units:
        label += f' [{calibration.units}]'

    return label


def domain_coloring(z, pure_phase=False):
    """
    Domain coloring function.

    Function to color a complex domain.

    Parameters
    ----------
    z : ndarray, complex
        Complex number to be colored.
    saturation : float, optional
        RGB color saturation. Default is 1.0.
    k : float, optional
        Scaling factor for the coloring. Default is 0.5.
    """

    phase = (np.angle(z) + np.pi) / (2 * np.pi)

    cmap = ListedColormap(hsluv)
    colors = cmap(phase)[..., :3]
    if not pure_phase:
        abs_z = np.abs(z)
        abs_z = (abs_z - abs_z.min()) / abs_z.ptp()
        colors = colors * abs_z[..., None]

    return colors


def _line_intersect_rectangle(point0, point1, lower_corner, upper_corner):
    if point0[0] == point1[0]:
        return (point0[0], lower_corner[1]), (point0[0], upper_corner[1])

    m = (point1[1] - point0[1]) / (point1[0] - point0[0])

    def y(x):
        return m * (x - point0[0]) + point0[1]

    def x(y):
        return (y - point0[1]) / m + point0[0]

    if y(0) < lower_corner[1]:
        intersect0 = (x(lower_corner[1]), y(x(lower_corner[1])))
    else:
        intersect0 = (0, y(lower_corner[0]))

    if y(upper_corner[0]) > upper_corner[1]:
        intersect1 = (x(upper_corner[1]), y(x(upper_corner[1])))
    else:
        intersect1 = (upper_corner[0], y(upper_corner[0]))

    return intersect0, intersect1



import numpy as np
from ase.io import read

from abtem.detect import AnnularDetector, FlexibleAnnularDetector, PixelatedDetector
from abtem.potentials import Potential
from abtem.scan import GridScan
from abtem.waves import Probe


def test_detector_consistency():
    atoms = read('data/srtio3_100.cif')
    atoms *= (4, 4, 1)

    potential = Potential(atoms, gpts=256, projection='infinite', slice_thickness=.5,
                          parametrization='kirkland', ).build(pbar=False)

    probe = Probe(energy=300e3, semiangle_cutoff=9.4, rolloff=0.05)

    flexible_detector = FlexibleAnnularDetector()
    annular_detector = AnnularDetector(inner=20, outer=40)
    pixelated_detector = PixelatedDetector()

    end = (potential.extent[0] / 4, potential.extent[1] / 4)

    gridscan = GridScan(start=[0, 0], end=end, sampling=.5)

    measurements = probe.scan(gridscan, [flexible_detector, pixelated_detector, annular_detector], potential,
                              pbar=False)

    assert np.allclose(measurements[0].integrate(20, 40).array, measurements[2].array)
    assert np.allclose(annular_detector.integrate(measurements[1]).array,
                       measurements[2].array)


def test_pixelated_detector():
    gpts = (512, 512)
    extent = (12, 12)
    probe = Probe(energy=60e3, extent=extent, gpts=gpts, semiangle_cutoff=80, rolloff=0.2)
    detector = PixelatedDetector(max_angle=30, resample='uniform')

    wave = probe.build().downsample(max_angle=30)
    measurement = detector.detect(wave)
    assert measurement.shape == wave.array.shape

    detector = PixelatedDetector(max_angle='valid', resample='uniform')
    measurement = detector.detect(wave)
    assert measurement.shape == wave.array.shape

    detector = PixelatedDetector(max_angle='limit', resample='uniform')
    measurement = detector.detect(wave)
    assert measurement.shape == wave.array.shape

    gpts = (512, 512)
    extent = (10, 12)

    probe = Probe(energy=60e3, extent=extent, gpts=gpts, semiangle_cutoff=80, rolloff=0.2)
    detector = PixelatedDetector(max_angle='valid', resample='uniform')

    wave = probe.build()
    measurement = detector.allocate_measurement(wave)
    assert measurement.array.shape[0] == measurement.array.shape[1]


def test_segmented_detector():
    # TODO : fill in test
    pass


import numpy as np
from ase import Atoms

from abtem import SMatrix, Potential, GridScan
from abtem.ionization import TransitionPotential, EELSDetector


class DummyProjectedAtomicTransition:

    def build(self, x):
        return 1


class DummyCarbonTransitions:
    Z = 6

    def get_transition_potentials(self, *arg, **kwargs):
        return [DummyProjectedAtomicTransition()]

    def __len__(self):
        return 1


# def test_dummy_transition():
#     atoms = Atoms('C', positions=[(2.5, 2.5, 1)], cell=(5, 5, 2))
#
#     transitions = [DummyCarbonTransitions()]
#
#     potential = Potential(atoms[[False]], slice_thickness=1, projection='infinite', parametrization='kirkland')
#
#     S = SMatrix(energy=100e3, sampling=0.02, semiangle_cutoff=20, rolloff=0.0)
#
#     S.match_grid(potential)
#
#     detector_eels = EELSDetector(collection_angle=20, interpolation=1)
#
#     transition_potential = TransitionPotential(transitions, atoms=atoms)
#
#     scan = GridScan((0, 0), potential.extent, sampling=0.9 * S.ctf.nyquist_sampling)
#
#     measurement_eels = S.coreloss_scan(scan,
#                                        detector_eels,
#                                        potential,
#                                        transition_potential, pbar=False)
#
#     assert np.allclose(measurement_eels.array, 1)


def test_tranpose():
    pass
    # from abtem.waves import FresnelPropagator, PlaneWave
    #
    # atoms = Atoms('O', positions=[(2, 2, 2)], cell=(4, 4, 4))
    #
    # potential = Potential(atoms, gpts=512, slice_thickness=4, projection='infinite', parametrization='kirkland')
    # potential = potential.build(pbar=False).as_transmission_function(100e3)
    #
    # wave = PlaneWave(energy=100e3, device='cpu')
    # wave.match_grid(potential)
    # wave = wave.build()
    #
    # # plt.imshow((potential.array * np.conjugate(potential.array))[0].real)
    # # plt.colorbar()
    #
    # wave = potential.transmit(wave)
    # wave = propagator.propagate(wave, potential_slice.thickness)
    # wave = potential.transmit(wave)
    # wave = propagator.propagate(wave, potential_slice.thickness)
    #
    # plt.imshow(np.abs(wave.array[0]) ** 2)
    # plt.colorbar()
    # plt.show()
    #
    # wave = propagator.propagate(wave, -potential_slice.thickness)
    # wave = potential.transmit(wave, conjugate=True)
    # wave = propagator.propagate(wave, -potential_slice.thickness)
    # wave = potential.transmit(wave, conjugate=True)




import numpy as np
import pytest

from abtem.base_classes import Accelerator
from abtem.utils import energy2mass, energy2wavelength, energy2sigma


def test_energy2mass():
    assert np.isclose(energy2mass(300e3), 1.445736928082275e-30)


def test_energy2wavelength():
    assert np.isclose(energy2wavelength(300e3), 0.01968748889772767)


def test_energy2sigma():
    assert np.isclose(energy2sigma(300e3), 0.0006526161464700888)


def test_energy():
    energy = Accelerator(energy=300e3)

    assert energy.energy == 300e3
    assert np.isclose(energy.wavelength, energy2wavelength(300e3))

    energy.energy = 200e3

    assert np.isclose(energy.wavelength, energy2wavelength(200e3))


def test_energy_raises():
    accelerator1 = Accelerator(300e3)
    accelerator2 = Accelerator()

    with pytest.raises(RuntimeError):
        accelerator2.check_is_defined()

    accelerator2.energy = 200e3
    with pytest.raises(RuntimeError):
        accelerator1.check_match(accelerator2)

    accelerator2.energy = accelerator1.energy
    accelerator1.check_match(accelerator2)


def test_accelerator_event():
    accelerator = Accelerator(300e3)

    accelerator.energy = 200e3
    assert accelerator.event._notify_count == 1



"""Module for describing different types of scans."""
from abc import ABCMeta, abstractmethod
from copy import copy
from typing import Union, Sequence, Tuple, List

import h5py
import numpy as np
from matplotlib.patches import Rectangle

from abtem.base_classes import Grid, HasGridMixin
from abtem.device import asnumpy
from abtem.measure import Calibration, Measurement
from abtem.utils import subdivide_into_batches, ProgressBar
from ase import Atom


class AbstractScan(metaclass=ABCMeta):
    """Abstract class to describe scans."""

    def __init__(self):
        self._batches = None

    def __len__(self):
        return self.num_positions

    @property
    def num_positions(self):
        return len(self.get_positions())

    @property
    @abstractmethod
    def shape(self) -> tuple:
        """The shape the scan."""
        pass

    @property
    @abstractmethod
    def calibrations(self) -> tuple:
        """The measurement calibrations associated with the scan."""
        pass

    @abstractmethod
    def get_positions(self):
        """Get the scan positions as numpy array."""
        pass

    @abstractmethod
    def insert_new_measurement(self, measurement, indices, new_values):
        """
        Insert new measurement values into a Measurement object or HDF5 file.

        Parameters
        ----------
        measurement : Measurement object
            The measurement to which new values are inserted.
        start : int
            First index of slice.
        end : int
            Last index of slice.
        new_values : ndarray
            New measurement values to be inserted. Length should be (end - start).
        """
        pass

    def generate_positions(self, max_batch, pbar=False):
        positions = self.get_positions()
        self._partition_batches(max_batch)

        if pbar:
            pbar = ProgressBar(total=len(self))

        for i in range(len(self._batches)):
            indices = self.get_next_batch()
            yield indices, positions[indices]

            if pbar:
                pbar.update(len(indices))

        if pbar:
            pbar.close()

    def get_next_batch(self):
        return self._batches.pop(0)

    def _partition_batches(self, max_batch):
        n = len(self)
        n_batches = (n + (-n % max_batch)) // max_batch
        batch_sizes = subdivide_into_batches(len(self), n_batches)

        self._batches = []

        start = 0
        for batch_size in batch_sizes:
            end = start + batch_size
            indices = np.arange(start, end, dtype=np.int)
            start += batch_size
            self._batches.append(indices)

    @abstractmethod
    def __copy__(self):
        pass

    def copy(self):
        """Make a copy."""
        return copy(self)


class PositionScan(AbstractScan):
    """
    Position scan object.

    Defines a scan based on user-provided positions.

    Parameters
    ----------
    positions : list
        A list of xy scan positions [Å].
    """

    def __init__(self, positions: np.ndarray):

        self._positions = np.array(positions)

        if (len(self._positions.shape) != 2) or (self._positions.shape[1] != 2):
            raise RuntimeError('The shape of the sequence of positions must be (n, 2).')

        super().__init__()

    @property
    def shape(self) -> tuple:
        return len(self),

    @property
    def calibrations(self) -> tuple:
        return None,

    def insert_new_measurement(self, measurement, indices, new_measurement):
        if isinstance(measurement, str):
            with h5py.File(measurement, 'a') as f:
                f['array'][indices] = asnumpy(new_measurement)

        else:
            measurement.array[indices] = asnumpy(new_measurement)

    def get_positions(self):
        return self._positions

    def add_to_mpl_plot(self, ax, marker: str = 'o', color: str = 'r', **kwargs):
        """
        Add a visualization of the scan positions to a matplotlib plot.

        Parameters
        ----------
        ax: matplotlib Axes
            The axes of the matplotlib plot the visualization should be added to.
        marker: str, optional
            Style of scan position markers. Default is '-'.
        color: str, optional
            Color of the scan position markers. Default is 'r'.
        kwargs:
            Additional options for matplotlib.pyplot.plot as keyword arguments.
        """
        ax.plot(*self.get_positions().T, marker=marker, linestyle='', color=color, **kwargs)

    def __copy__(self):
        return self.__class__(self._positions.copy())


class LineScan(AbstractScan, HasGridMixin):
    """
    Line scan object.

    Defines a scan along a straight line.

    Parameters
    ----------
    start : two float
        Start point of the scan [Å].
    end : two float
        End point of the scan [Å].
    gpts: int
        Number of scan positions.
    sampling: float
        Sampling rate of scan positions [1 / Å].
    endpoint: bool
        If True, end is the last position. Otherwise, it is not included. Default is True.
    """

    def __init__(self,
                 start: Union[Tuple[float, float], Atom],
                 end: Union[Tuple[float, float], Atom] = None,
                 angle: float = 0.,
                 gpts: int = None,
                 sampling: float = None,
                 margin: float = 0.,
                 endpoint: bool = True):

        super().__init__()

        if isinstance(start, Atom):
            start = (start.x, start.y)

        if isinstance(end, Atom):
            end = (end.x, end.y)

        # if (end is not None) & (angle is not None):
        #    raise ValueError('only one of "end" and "angle" may be specified')

        if (gpts is None) & (sampling is None):
            raise RuntimeError('grid gpts or sampling must be set')

        self._grid = Grid(gpts=gpts, sampling=sampling, endpoint=endpoint, dimensions=1)

        self._start = start[:2]
        self._margin = margin

        if end is not None:
            self._set_direction_and_extent(self._start, end[:2])
        else:
            self.angle = angle
            self.extent = 2 * self._margin

    def _set_direction_and_extent(self, start: Tuple[float, float], end: Tuple[float, float]):
        difference = np.array(end) - np.array(start)
        extent = np.linalg.norm(difference, axis=0)
        self._direction = difference / extent
        extent = extent + 2 * self._margin
        if extent == 0.:
            raise RuntimeError('scan has no extent')
        self.extent = extent

    @property
    def shape(self) -> Tuple[int]:
        return self.gpts[0],

    @property
    def calibrations(self) -> Tuple[Calibration]:
        return Calibration(offset=0, sampling=self.sampling[0], units='Å', name='x', endpoint=self.grid.endpoint[0]),

    @property
    def start(self) -> Tuple[float, float]:
        """
        Start point of the scan [Å].
        """
        return self._start

    @start.setter
    def start(self, start: Tuple[float, float]):
        self._start = start
        self._set_direction_and_extent(self._start, self.end)

    @property
    def end(self) -> Tuple[float, float]:
        """
        End point of the scan [Å].
        """
        return (self.start[0] + self.direction[0] * self.extent[0] - self.direction[0] * 2 * self._margin,
                self.start[1] + self.direction[1] * self.extent[0] - self.direction[1] * 2 * self._margin)

    @end.setter
    def end(self, end: Tuple[float, float]):
        self._set_direction_and_extent(self.start, end)

    @property
    def angle(self) -> float:
        """
        End point of the scan [Å].
        """
        return np.arctan2(self._direction[0], self._direction[1])

    @angle.setter
    def angle(self, angle: float):
        self._direction = (np.cos(np.deg2rad(angle)), np.sin(np.deg2rad(angle)))

    @property
    def direction(self) -> Tuple[float, float]:
        """Direction of the scan line."""
        return self._direction

    @property
    def margin(self) -> float:
        return self._margin

    def insert_new_measurement(self,
                               measurement: Measurement,
                               indices,
                               new_measurement_values: np.ndarray):

        if isinstance(measurement, str):
            with h5py.File(measurement, 'a') as f:
                f['array'][indices] += asnumpy(new_measurement_values)

        else:
            measurement.array[indices] += asnumpy(new_measurement_values)

    @property
    def margin_start(self) -> Tuple[float, float]:
        return self.start[0] - self.direction[0] * self.margin, self.start[1] - self.direction[1] * self.margin

    @property
    def margin_end(self) -> Tuple[float, float]:
        return self.end[0] + self.direction[0] * self.margin, self.end[1] + self.direction[1] * self.margin

    def get_positions(self) -> np.ndarray:
        start = self.margin_start
        end = self.margin_end
        x = np.linspace(start[0], end[0], self.gpts[0], endpoint=self.grid.endpoint[0])
        y = np.linspace(start[1], end[1], self.gpts[0], endpoint=self.grid.endpoint[0])
        return np.stack((np.reshape(x, (-1,)), np.reshape(y, (-1,))), axis=1)

    def add_to_mpl_plot(self, ax, color: str = 'r', width: float = 0.2, **kwargs):
        """
        Add a visualization of a scan line to a matplotlib plot.

        Parameters
        ----------
        ax : matplotlib Axes
            The axes of the matplotlib plot the visualization should be added to.
        linestyle : str, optional
            Linestyle of scan line. Default is '-'.
        color : str, optional
            Color of the scan line. Default is 'r'.
        kwargs :
            Additional options for matplotlib.pyplot.plot as keyword arguments.
        """
        start = self.margin_start
        end = self.margin_end

        from matplotlib.lines import Line2D

        class LineDataUnits(Line2D):
            def __init__(self, *args, **kwargs):
                _lw_data = kwargs.pop("linewidth", 1)
                super().__init__(*args, **kwargs)
                self._lw_data = _lw_data

            def _get_lw(self):
                if self.axes is not None:
                    ppd = 72. / self.axes.figure.dpi
                    trans = self.axes.transData.transform
                    return ((trans((1, self._lw_data)) - trans((0, 0))) * ppd)[1]
                else:
                    return 1

            def _set_lw(self, lw):
                self._lw_data = lw

            _linewidth = property(_get_lw, _set_lw)

        line = LineDataUnits([start[0], end[0]], [start[1], end[1]], linewidth=width, color=color, **kwargs)
        ax.add_line(line)

        # ax.plot([start[0], end[0]], [start[1], end[1]], linestyle=linestyle, color=color, **kwargs)

    def __copy__(self):
        return self.__class__(start=self.start, end=self.end, gpts=self.gpts, endpoint=self.grid.endpoint[0])


class GridScan(AbstractScan, HasGridMixin):
    """
    Grid scan object.

    Defines a scan on a regular grid.

    Parameters
    ----------
    start : two float
        Start corner of the scan [Å].
    end : two float
        End corner of the scan [Å].
    gpts : two int
        Number of scan positions in the x- and y-direction of the scan.
    sampling : two float
        Sampling rate of scan positions [1 / Å].
    endpoint : bool
        If True, end is the last position. Otherwise, it is not included. Default is False.
    batch_partition : 'squares' or 'lines'
        Specify how to split the scan into batches. If 'squares', the scan position batches are divided into the best
        matching squares for the batch size. If 'lines', the batches are divided into lines of scan positions.
    measurement_shift : two int
        The insertion indices of new measurements will be shifted by this amount in x and y. This is used for
        correctly inserting measurements collected from a partitioned scan.
    """

    def __init__(self,
                 start: Sequence[float],
                 end: Sequence[float],
                 gpts: Union[int, Sequence[int]] = None,
                 sampling: Union[float, Sequence[float]] = None,
                 endpoint: bool = False,
                 batch_partition: str = 'squares',
                 measurement_shift: Sequence[int] = None):

        super().__init__()

        try:
            self._start = np.array(start)[:2]
            end = np.array(end)[:2]
            assert (self._start.shape == (2,)) & (end.shape == (2,))
        except:
            raise ValueError('Scan start/end has incorrect shape')

        if (gpts is None) & (sampling is None):
            raise RuntimeError('Grid gpts or sampling must be set')

        if not batch_partition.lower() in ['squares', 'lines']:
            raise ValueError('batch partition must be "squares" or "lines"')

        self._batch_partition = batch_partition
        self._measurement_shift = measurement_shift

        self._grid = Grid(extent=end - start, gpts=gpts, sampling=sampling, dimensions=2, endpoint=endpoint)

    @property
    def shape(self):
        return self.gpts

    @property
    def calibrations(self) -> tuple:
        return (Calibration(offset=self.start[0], sampling=self.sampling[0], units='Å', name='x',
                            endpoint=self.grid.endpoint[0]),
                Calibration(offset=self.start[1], sampling=self.sampling[1], units='Å', name='y',
                            endpoint=self.grid.endpoint[1]))

    @property
    def start(self) -> np.ndarray:
        """Start corner of the scan [Å]."""
        return self._start

    @start.setter
    def start(self, start: Sequence[float]):
        self._start = np.array(start)
        self.extent = self.end - self._start

    @property
    def end(self) -> np.ndarray:
        """End corner of the scan [Å]."""
        return self.start + self.extent

    @end.setter
    def end(self, end: Sequence[float]):
        self.extent = np.array(end) - self.start

    def get_scan_area(self) -> float:
        """Get the area of the scan."""
        height = abs(self.start[0] - self.end[0])
        width = abs(self.start[1] - self.end[1])
        return height * width

    def get_positions(self) -> np.ndarray:
        x = np.linspace(self.start[0], self.end[0], self.gpts[0], endpoint=self.grid.endpoint[0])
        y = np.linspace(self.start[1], self.end[1], self.gpts[1], endpoint=self.grid.endpoint[1])
        x, y = np.meshgrid(x, y, indexing='ij')
        return np.stack((np.reshape(x, (-1,)),
                         np.reshape(y, (-1,))), axis=1)

    def insert_new_measurement(self, measurement, indices: np.ndarray, new_measurement: np.ndarray):
        x, y = np.unravel_index(indices, self.shape)

        if self._measurement_shift is not None:
            x += self._measurement_shift[0]
            y += self._measurement_shift[1]

        if isinstance(measurement, str):
            with h5py.File(measurement, 'a') as f:
                for unique in np.unique(x):
                    f['array'][unique, y[unique == x]] += asnumpy(new_measurement[unique == x])
        else:
            measurement.array[x, y] += asnumpy(new_measurement)

    def partition_scan(self, partitions: Sequence[int]) -> List['GridScan']:
        """
        Partition the scan into smaller grid scans

        Parameters
        ----------
        partitions : two int
            The number of partitions to create in x and y.

        Returns
        -------
        List of GridScan objects
        """
        Nx = subdivide_into_batches(self.gpts[0], partitions[0])
        Ny = subdivide_into_batches(self.gpts[1], partitions[1])
        Sx = np.concatenate(([0], np.cumsum(Nx)))
        Sy = np.concatenate(([0], np.cumsum(Ny)))

        scans = []
        for i, nx in enumerate(Nx):
            for j, ny in enumerate(Ny):
                start = [Sx[i] * self.sampling[0], Sy[j] * self.sampling[1]]
                end = [start[0] + nx * self.sampling[0], start[1] + ny * self.sampling[1]]
                endpoint = [False, False]

                if i + 1 == partitions[0]:
                    endpoint[0] = self.grid.endpoint[0]
                    if endpoint[0]:
                        end[0] -= self.sampling[0]

                if j + 1 == partitions[1]:
                    endpoint[1] = self.grid.endpoint[1]
                    if endpoint[1]:
                        end[1] -= self.sampling[1]

                scan = self.__class__(start,
                                      end,
                                      gpts=(nx, ny),
                                      endpoint=endpoint,
                                      batch_partition='squares',
                                      measurement_shift=(Sx[i], Sy[j]))

                scans.append(scan)
        return scans

    def _partition_batches(self, max_batch: int):
        if self._batch_partition == 'lines':
            super()._partition_batches(max_batch)
            return

        if max_batch == 1:
            self._batches = [[i] for i in range(len(self))]
            return

        max_batch_x = int(np.floor(np.sqrt(max_batch)))
        max_batch_y = int(np.floor(np.sqrt(max_batch)))

        Nx = subdivide_into_batches(self.gpts[0], (self.gpts[0] + (-self.gpts[0] % max_batch_x)) // max_batch_x)
        Ny = subdivide_into_batches(self.gpts[1], (self.gpts[1] + (-self.gpts[1] % max_batch_y)) // max_batch_y)

        self._batches = []
        Sx = np.concatenate(([0], np.cumsum(Nx)))
        Sy = np.concatenate(([0], np.cumsum(Ny)))

        for i, nx in enumerate(Nx):
            for j, ny in enumerate(Ny):
                x = np.arange(Sx[i], Sx[i] + nx, dtype=np.int)
                y = np.arange(Sy[j], Sy[j] + ny, dtype=np.int)
                self._batches.append((y[None] + x[:, None] * self.gpts[1]).ravel())

    def add_to_mpl_plot(self, ax, alpha: float = .33, facecolor: str = 'r', edgecolor: str = 'r', **kwargs):
        """
        Add a visualization of the scan area to a matplotlib plot.

        Parameters
        ----------
        ax : matplotlib Axes
            The axes of the matplotlib plot the visualization should be added to.
        alpha : float, optional
            Transparency of the scan area visualization. Default is 0.33.
        facecolor : str, optional
            Color of the scan area visualization.
        edgecolor : str, optional
            Color of the edge of the scan area visualization.
        kwargs :
            Additional options for matplotlib.patches.Rectangle used for scan area visualization as keyword arguments.
        """
        rect = Rectangle(tuple(self.start), *self.extent, alpha=alpha, facecolor=facecolor, edgecolor=edgecolor,
                         **kwargs)
        ax.add_patch(rect)

    def __copy__(self):
        return self.__class__(start=self.start,
                              end=self.end,
                              gpts=self.gpts,
                              endpoint=self.grid.endpoint,
                              batch_partition=self._batch_partition,
                              measurement_shift=self._measurement_shift)



import numpy as np
from scipy.spatial import cKDTree, ConvexHull


def triangle_area(pt1, pt2, pt3):
    a = pt1[0] * pt2[1] - pt2[0] * pt1[1] + pt2[0] * pt3[1] - pt3[0] * pt2[1] + pt3[0] * pt1[1] - pt1[0] * pt3[1]
    return abs(a) / 2


def circumcircle_radius(pt0, pt1, pt2):
    a = np.linalg.norm(pt0 - pt1)
    b = np.linalg.norm(pt1 - pt2)
    c = np.linalg.norm(pt2 - pt0)

    t_area = triangle_area(pt0, pt1, pt2)

    if t_area > 0:
        radius = (a * b * c) / (4 * t_area)
    else:
        radius = np.nan

    return radius


def circumcenter(pt0, pt1, pt2, eps=1e-12):
    a_x, a_y = pt0
    b_x, b_y = pt1
    c_x, c_y = pt2

    bc_y_diff = b_y - c_y
    ca_y_diff = c_y - a_y
    ab_y_diff = a_y - b_y
    cb_x_diff = c_x - b_x
    ac_x_diff = a_x - c_x
    ba_x_diff = b_x - a_x

    d_div = (a_x * bc_y_diff + b_x * ca_y_diff + c_x * ab_y_diff)

    if abs(d_div) < eps:
        line = np.array([pt0, pt1, pt2])
        distances = np.sum((line[..., None] - line.T[None]) ** 2, axis=1)
        line[np.where(distances == distances.max())[0]].mean(axis=0)
        return np.mean([pt0, pt1, pt2], axis=0)

    d_inv = 0.5 / d_div

    a_mag = a_x * a_x + a_y * a_y
    b_mag = b_x * b_x + b_y * b_y
    c_mag = c_x * c_x + c_y * c_y

    cx = (a_mag * bc_y_diff + b_mag * ca_y_diff + c_mag * ab_y_diff) * d_inv
    cy = (a_mag * cb_x_diff + b_mag * ac_x_diff + c_mag * ba_x_diff) * d_inv
    return cx, cy


def find_natural_neighbors(tri, grid_points):
    tree = cKDTree(grid_points)

    # in_triangulation = tri.find_simplex(tree.data) >= 0

    triangle_info = []
    members = {key: [] for key in range(len(tree.data))}
    for i, indices in enumerate(tri.simplices):
        triangle = tri.points[indices]
        cc = circumcenter(*triangle)
        r = circumcircle_radius(*triangle)
        triangle_info.append(cc)

        for point in tree.query_ball_point(cc, r):
            members[point].append(i)

    return members, np.array(triangle_info)


def find_local_boundary(tri, triangles):
    edges = []

    for triangle in triangles:
        for i in range(3):

            pt1 = tri.simplices[triangle][i]
            pt2 = tri.simplices[triangle][(i + 1) % 3]

            if (pt1, pt2) in edges:
                edges.remove((pt1, pt2))

            elif (pt2, pt1) in edges:
                edges.remove((pt2, pt1))

            else:
                edges.append((pt1, pt2))

    return edges


def polygon_area(poly):
    a = 0.0
    n = len(poly)

    for i in range(n):
        a += poly[i][0] * poly[(i + 1) % n][1] - poly[(i + 1) % n][0] * poly[i][1]

    return abs(a) / 2.0


def order_edges(edges):
    edge = edges[0]
    edges = edges[1:]

    ordered_edges = [edge]

    num_max = len(edges)
    while len(edges) > 0 and num_max > 0:

        match = edge[1]

        for search_edge in edges:
            vertex = search_edge[0]
            if match == vertex:
                edge = search_edge
                edges.remove(edge)
                ordered_edges.append(search_edge)
                break
        num_max -= 1

    return ordered_edges


def natural_neighbor_weights(points, query_point, tri, neighbors, circumcenters):
    weights = np.zeros(len(points))

    overlap = np.isclose(query_point[0], points[:, 0]) * np.isclose(query_point[1], points[:, 1])

    if np.any(overlap):
        weights[np.where(overlap)[0]] = 1.
        return weights

    edges = find_local_boundary(tri, neighbors)
    edge_vertices = [segment[0] for segment in order_edges(edges)]
    num_vertices = len(edge_vertices)

    p1 = edge_vertices[0]
    p2 = edge_vertices[1]

    c1 = circumcenter(query_point, tri.points[p1], tri.points[p2])
    polygon = [c1]
    for i in range(num_vertices):

        p3 = edge_vertices[(i + 2) % num_vertices]

        c2 = circumcenter(query_point, tri.points[p3], tri.points[p2])
        polygon.append(c2)

        for check_tri in neighbors:
            if p2 in tri.simplices[check_tri]:
                polygon.append(circumcenters[check_tri])

        pts = [polygon[i] for i in ConvexHull(polygon).vertices]
        area = polygon_area(pts)
        weights[(tri.points[p2][0] == points[:, 0]) & (tri.points[p2][1] == points[:, 1])] += area

        polygon = [c2]
        p2 = p3

    return weights / weights.sum()



import os

import numpy as np
import pytest
from ase.io import read

from abtem.potentials import Potential
from abtem.structures import orthogonalize_cell
from ase import Atoms
from ase import units


@pytest.mark.gpaw
def test_dft():
    from gpaw import GPAW
    from abtem.dft import GPAWPotential

    atoms = read(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data/hexagonal_graphene.cif'))

    gpaw = GPAW(h=.1, txt=None, kpts=(3, 3, 1))
    atoms.calc = gpaw
    atoms.get_potential_energy()

    dft_pot = GPAWPotential(gpaw, sampling=.02)

    dft_array = dft_pot.build()

    dft_potential = dft_array.tile((3, 2))

    atoms = orthogonalize_cell(gpaw.atoms) * (3, 2, 1)

    iam_potential = Potential(atoms, gpts=dft_potential.gpts, cutoff_tolerance=1e-4, device='cpu').build()

    projected_iam = iam_potential.array.sum(0)
    projected_iam -= projected_iam.min()

    projected_dft = dft_potential.array.sum(0)
    projected_dft -= projected_dft.min()

    absolute_difference = projected_iam - projected_dft

    valid = np.abs(projected_iam) > 1
    relative_difference = np.zeros_like(projected_iam)
    relative_difference[:] = np.nan
    relative_difference[valid] = 100 * (projected_iam[valid] - projected_dft[valid]) / projected_iam[valid]

    assert np.isclose(9.553661, absolute_difference.max(), atol=.1)
    assert np.isclose(43.573837, relative_difference[valid].max(), atol=.1)


@pytest.mark.gpaw
def test_compare_abtem_to_gpaw():
    from gpaw import GPAW
    from gpaw.utilities.ps2ae import PS2AE
    from abtem.dft import GPAWPotential

    atoms = Atoms('C', positions=[(0, 1, 2)], cell=(2, 2, 4), pbc=True)

    calc = GPAW(h=.2, txt=None, kpts=(2, 2, 1))
    atoms.calc = calc
    atoms.get_potential_energy()

    h = 0.01
    t = PS2AE(calc, h=h)
    ae = (-t.get_electrostatic_potential(rcgauss=.02 * units.Bohr) * h).sum(-1)
    ae -= ae.min()

    dft_pot = GPAWPotential(calc, gpts=ae.shape, core_size=.02, slice_thickness=4)
    dft_array = dft_pot.build(pbar=False)
    abtem_ae = dft_array.array.sum(0)
    abtem_ae -= abtem_ae.min()

    valid = abtem_ae > 1

    assert np.all(((abtem_ae[valid] - ae[valid]) / ae[valid]).max() < .001)



import numpy as np

from abtem.detect import PixelatedDetector
from abtem.measure import Measurement, calibrations_from_grid
from abtem.potentials import Potential, PotentialArray
from abtem.scan import LineScan, GridScan
from abtem.waves import Probe, Waves


def test_export_import_potential(tmp_path, graphene_atoms):
    d = tmp_path / 'sub'
    d.mkdir()
    path = d / 'potential.hdf5'

    potential = Potential(graphene_atoms, sampling=.05)
    precalculated_potential = potential.build(pbar=False)
    precalculated_potential.write(path)
    imported_potential = PotentialArray.read(path)
    assert np.allclose(imported_potential.array, precalculated_potential.array)
    assert np.allclose(imported_potential.extent, precalculated_potential.extent)
    assert np.allclose(imported_potential._slice_thicknesses, precalculated_potential._slice_thicknesses)


def test_export_import_waves(tmp_path):
    d = tmp_path / 'sub'
    d.mkdir()
    path = d / 'waves.hdf5'

    waves = Probe(semiangle_cutoff=30, sampling=.05, extent=10, energy=80e3).build()
    waves.write(path)
    imported_waves = Waves.read(path)
    assert np.allclose(waves.array, imported_waves.array)
    assert np.allclose(waves.extent, imported_waves.extent)
    assert np.allclose(waves.energy, imported_waves.energy)


def test_export_import_measurement(tmp_path):
    d = tmp_path / 'sub'
    d.mkdir()
    path = d / 'measurement.hdf5'

    calibrations = calibrations_from_grid((512, 256), (.1, .3), ['x', 'y'], 'Å')

    measurement = Measurement(np.random.rand(512, 256), calibrations)
    measurement.write(path)
    imported_measurement = Measurement.read(path)
    assert np.allclose(measurement.array, imported_measurement.array)
    assert measurement.calibrations[0] == imported_measurement.calibrations[0]
    assert measurement.calibrations[1] == imported_measurement.calibrations[1]


def test_linescan_to_file(tmp_path, graphene_atoms):
    d = tmp_path / 'sub'
    d.mkdir()
    path = d / 'measurement2.hdf5'

    potential = Potential(atoms=graphene_atoms, sampling=.05)

    probe = Probe(energy=200e3, semiangle_cutoff=30)

    probe.grid.match(potential)

    scan = LineScan(start=[0, 0], end=[0, potential.extent[1]], gpts=20)

    detector = PixelatedDetector()
    export_detector = PixelatedDetector(save_file=path)

    measurements = probe.scan(scan, [detector, export_detector], potential, pbar=False)

    measurement = measurements[0]
    imported_measurement = Measurement.read(measurements[1])

    assert np.allclose(measurement.array, imported_measurement.array)
    assert measurement.calibrations[0] == imported_measurement.calibrations[0]
    assert measurement.calibrations[1] == imported_measurement.calibrations[1]


def test_gridscan_to_file(tmp_path, graphene_atoms):
    d = tmp_path / 'sub'
    d.mkdir()
    path = d / 'measurement2.hdf5'

    potential = Potential(atoms=graphene_atoms, sampling=.05)

    probe = Probe(energy=200e3, semiangle_cutoff=30)

    probe.grid.match(potential)

    scan = GridScan(start=[0, 0], end=[0, potential.extent[1]], gpts=(10, 9))

    detector = PixelatedDetector()
    export_detector = PixelatedDetector(save_file=path)

    measurements = probe.scan(scan, [detector, export_detector], potential, pbar=False)

    measurement = measurements[0]
    imported_measurement = Measurement.read(measurements[1])

    assert np.allclose(measurement.array, imported_measurement.array)
    assert measurement.calibrations[0] == imported_measurement.calibrations[0]
    assert measurement.calibrations[1] == imported_measurement.calibrations[1]



"""Main abTEM module."""
from abtem import waves
from abtem.detect import AnnularDetector, FlexibleAnnularDetector, SegmentedDetector, PixelatedDetector, \
    WavefunctionDetector
from abtem.visualize.mpl import show_atoms
from abtem.potentials import Potential, PotentialArray
from abtem.scan import GridScan, LineScan
from abtem.temperature import AbstractFrozenPhonons, FrozenPhonons, MDFrozenPhonons
from abtem.transfer import CTF
from abtem.waves import PlaneWave, Probe, SMatrix, Waves, FresnelPropagator
from abtem.measure import Measurement

from abtem._version import __version__



"""Module to describe the contrast transfer function."""
from collections import defaultdict
from typing import Mapping, Union

import numpy as np
from abtem.base_classes import HasAcceleratorMixin, HasEventMixin, Accelerator, watched_method, watched_property, Event, \
    Grid
from abtem.device import get_array_module, get_device_function
from abtem.measure import Measurement, Calibration
from abtem.utils import energy2wavelength, spatial_frequencies, polar_coordinates

#: Symbols for the polar representation of all optical aberrations up to the fifth order.
polar_symbols = ('C10', 'C12', 'phi12',
                 'C21', 'phi21', 'C23', 'phi23',
                 'C30', 'C32', 'phi32', 'C34', 'phi34',
                 'C41', 'phi41', 'C43', 'phi43', 'C45', 'phi45',
                 'C50', 'C52', 'phi52', 'C54', 'phi54', 'C56', 'phi56')

#: Aliases for the most commonly used optical aberrations.
polar_aliases = {'defocus': 'C10', 'astigmatism': 'C12', 'astigmatism_angle': 'phi12',
                 'coma': 'C21', 'coma_angle': 'phi21',
                 'Cs': 'C30',
                 'C5': 'C50'}


class CTF(HasAcceleratorMixin, HasEventMixin):
    """
    Contrast transfer function object

    The Contrast Transfer Function (CTF) describes the aberrations of the objective lens in HRTEM and specifies how the
    condenser system shapes the probe in STEM.

    abTEM implements phase aberrations up to 5th order using polar coefficients. See Eq. 2.22 in the reference [1]_.
    Cartesian coefficients can be converted to polar using the utility function abtem.transfer.cartesian2polar.

    Partial coherence is included as an envelope in the quasi-coherent approximation. See Chapter 3.2 in reference [1]_.

    For a more detailed discussion with examples, see our `walkthrough
    <https://abtem.readthedocs.io/en/latest/walkthrough/05_contrast_transfer_function.html>`_.

    Parameters
    ----------
    semiangle_cutoff: float
        The semiangle cutoff describes the sharp Fourier space cutoff due to the objective aperture [mrad].
    rolloff: float
        Tapers the cutoff edge over the given angular range [mrad].
    focal_spread: float
        The 1/e width of the focal spread due to chromatic aberration and lens current instability [Å].
    angular_spread: float
        The 1/e width of the angular deviations due to source size [mrad].
    gaussian_spread:
        The 1/e width image deflections due to vibrations and thermal magnetic noise [Å].
    energy: float
        The electron energy of the wave functions this contrast transfer function will be applied to [eV].
    parameters: dict
        Mapping from aberration symbols to their corresponding values. All aberration magnitudes should be given in Å
        and angles should be given in radians.
    kwargs:
        Provide the aberration coefficients as keyword arguments.

    References
    ----------
    .. [1] Kirkland, E. J. (2010). Advanced Computing in Electron Microscopy (2nd ed.). Springer.

    """

    def __init__(self,
                 semiangle_cutoff: float = np.inf,
                 rolloff: float = 2,
                 focal_spread: float = 0.,
                 angular_spread: float = 0.,
                 gaussian_spread: float = 0.,
                 energy: float = None,
                 parameters: Mapping[str, float] = None,
                 aperture=None,
                 **kwargs):

        for key in kwargs.keys():
            if (key not in polar_symbols) and (key not in polar_aliases.keys()):
                raise ValueError('{} not a recognized parameter'.format(key))

        self._event = Event()

        self._accelerator = Accelerator(energy=energy)
        self._accelerator.observe(self.event.notify)

        self._semiangle_cutoff = semiangle_cutoff
        self._rolloff = rolloff
        self._focal_spread = focal_spread
        self._angular_spread = angular_spread
        self._gaussian_spread = gaussian_spread

        self._parameters = dict(zip(polar_symbols, [0.] * len(polar_symbols)))

        self._aperture = aperture

        if self._aperture is not None:
            self._aperture.accelerator.match(self)

        if parameters is None:
            parameters = {}

        parameters.update(kwargs)
        self.set_parameters(parameters)

        def parametrization_property(key):

            def getter(self):
                return self._parameters[key]

            def setter(self, value):
                old = getattr(self, key)
                self._parameters[key] = value
                self.event.notify({'notifier': self, 'name': key, 'change': old != value})

            return property(getter, setter)

        for symbol in polar_symbols:
            setattr(self.__class__, symbol, parametrization_property(symbol))

        for key, value in polar_aliases.items():
            if key != 'defocus':
                setattr(self.__class__, key, parametrization_property(value))

    @property
    def nyquist_sampling(self):
        return 1 / (4 * self.semiangle_cutoff / self.wavelength * 1e-3)

    @property
    def parameters(self):
        """The parameters."""
        return self._parameters

    @property
    def defocus(self) -> float:
        """The defocus [Å]."""
        return - self._parameters['C10']

    @defocus.setter
    def defocus(self, value: float):
        self.C10 = -value

    @property
    def semiangle_cutoff(self) -> float:
        """The semi-angle cutoff [mrad]."""
        return self._semiangle_cutoff

    @semiangle_cutoff.setter
    @watched_property('_event')
    def semiangle_cutoff(self, value: float):
        self._semiangle_cutoff = value

    @property
    def rolloff(self) -> float:
        """The fraction of soft tapering of the cutoff."""
        return self._rolloff

    @rolloff.setter
    @watched_property('_event')
    def rolloff(self, value: float):
        self._rolloff = value

    @property
    def focal_spread(self) -> float:
        """The focal spread [Å]."""
        return self._focal_spread

    @focal_spread.setter
    @watched_property('_event')
    def focal_spread(self, value: float):
        """The angular spread [mrad]."""
        self._focal_spread = value

    @property
    def angular_spread(self) -> float:
        return self._angular_spread

    @angular_spread.setter
    @watched_property('_event')
    def angular_spread(self, value: float):
        self._angular_spread = value

    @property
    def gaussian_spread(self) -> float:
        """The Gaussian spread [Å]."""
        return self._gaussian_spread

    @gaussian_spread.setter
    @watched_property('_event')
    def gaussian_spread(self, value: float):
        self._gaussian_spread = value

    @watched_method('_event')
    def set_parameters(self, parameters: dict):
        """
        Set the phase of the phase aberration.

        Parameters
        ----------
        parameters: dict
            Mapping from aberration symbols to their corresponding values.
        """

        for symbol, value in parameters.items():
            if symbol in self._parameters.keys():
                self._parameters[symbol] = value

            elif symbol == 'defocus':
                self._parameters[polar_aliases[symbol]] = -value

            elif symbol in polar_aliases.keys():
                self._parameters[polar_aliases[symbol]] = value

            else:
                raise ValueError('{} not a recognized parameter'.format(symbol))

        return parameters

    def evaluate_aperture(self,
                          alpha: Union[float, np.ndarray],
                          phi: Union[float, np.ndarray] = None) -> Union[float, np.ndarray]:

        if self._aperture is not None:
            return self._aperture.evaluate(alpha, phi)

        xp = get_array_module(alpha)
        semiangle_cutoff = self.semiangle_cutoff / 1000

        if self.semiangle_cutoff == xp.inf:
            return xp.ones_like(alpha)

        if self.rolloff > 0.:
            rolloff = self.rolloff / 1000.  # * semiangle_cutoff
            array = .5 * (1 + xp.cos(np.pi * (alpha - semiangle_cutoff + rolloff) / rolloff))
            array[alpha > semiangle_cutoff] = 0.
            array = xp.where(alpha > semiangle_cutoff - rolloff, array, xp.ones_like(alpha, dtype=xp.float32))
        else:
            array = xp.array(alpha < semiangle_cutoff).astype(xp.float32)
        return array

    def evaluate_temporal_envelope(self, alpha: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        xp = get_array_module(alpha)
        return xp.exp(- (.5 * xp.pi / self.wavelength * self.focal_spread * alpha ** 2) ** 2).astype(xp.float32)

    def evaluate_gaussian_envelope(self, alpha: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        xp = get_array_module(alpha)
        return xp.exp(- .5 * self.gaussian_spread ** 2 * alpha ** 2 / self.wavelength ** 2)

    def evaluate_spatial_envelope(self, alpha: Union[float, np.ndarray], phi: Union[float, np.ndarray]) -> \
            Union[float, np.ndarray]:
        xp = get_array_module(alpha)
        p = self.parameters
        dchi_dk = 2 * xp.pi / self.wavelength * (
                (p['C12'] * xp.cos(2. * (phi - p['phi12'])) + p['C10']) * alpha +
                (p['C23'] * xp.cos(3. * (phi - p['phi23'])) +
                 p['C21'] * xp.cos(1. * (phi - p['phi21']))) * alpha ** 2 +
                (p['C34'] * xp.cos(4. * (phi - p['phi34'])) +
                 p['C32'] * xp.cos(2. * (phi - p['phi32'])) + p['C30']) * alpha ** 3 +
                (p['C45'] * xp.cos(5. * (phi - p['phi45'])) +
                 p['C43'] * xp.cos(3. * (phi - p['phi43'])) +
                 p['C41'] * xp.cos(1. * (phi - p['phi41']))) * alpha ** 4 +
                (p['C56'] * xp.cos(6. * (phi - p['phi56'])) +
                 p['C54'] * xp.cos(4. * (phi - p['phi54'])) +
                 p['C52'] * xp.cos(2. * (phi - p['phi52'])) + p['C50']) * alpha ** 5)

        dchi_dphi = -2 * xp.pi / self.wavelength * (
                1 / 2. * (2. * p['C12'] * xp.sin(2. * (phi - p['phi12']))) * alpha +
                1 / 3. * (3. * p['C23'] * xp.sin(3. * (phi - p['phi23'])) +
                          1. * p['C21'] * xp.sin(1. * (phi - p['phi21']))) * alpha ** 2 +
                1 / 4. * (4. * p['C34'] * xp.sin(4. * (phi - p['phi34'])) +
                          2. * p['C32'] * xp.sin(2. * (phi - p['phi32']))) * alpha ** 3 +
                1 / 5. * (5. * p['C45'] * xp.sin(5. * (phi - p['phi45'])) +
                          3. * p['C43'] * xp.sin(3. * (phi - p['phi43'])) +
                          1. * p['C41'] * xp.sin(1. * (phi - p['phi41']))) * alpha ** 4 +
                1 / 6. * (6. * p['C56'] * xp.sin(6. * (phi - p['phi56'])) +
                          4. * p['C54'] * xp.sin(4. * (phi - p['phi54'])) +
                          2. * p['C52'] * xp.sin(2. * (phi - p['phi52']))) * alpha ** 5)

        return xp.exp(-xp.sign(self.angular_spread) * (self.angular_spread / 2 / 1000) ** 2 *
                      (dchi_dk ** 2 + dchi_dphi ** 2))

    def evaluate_chi(self, alpha: Union[float, np.ndarray], phi: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        xp = get_array_module(alpha)
        p = self.parameters

        alpha2 = alpha ** 2
        alpha = xp.array(alpha)

        array = xp.zeros(alpha.shape, dtype=np.float32)
        if any([p[symbol] != 0. for symbol in ('C10', 'C12', 'phi12')]):
            array += (1 / 2 * alpha2 *
                      (p['C10'] +
                       p['C12'] * xp.cos(2 * (phi - p['phi12']))))

        if any([p[symbol] != 0. for symbol in ('C21', 'phi21', 'C23', 'phi23')]):
            array += (1 / 3 * alpha2 * alpha *
                      (p['C21'] * xp.cos(phi - p['phi21']) +
                       p['C23'] * xp.cos(3 * (phi - p['phi23']))))

        if any([p[symbol] != 0. for symbol in ('C30', 'C32', 'phi32', 'C34', 'phi34')]):
            array += (1 / 4 * alpha2 ** 2 *
                      (p['C30'] +
                       p['C32'] * xp.cos(2 * (phi - p['phi32'])) +
                       p['C34'] * xp.cos(4 * (phi - p['phi34']))))

        if any([p[symbol] != 0. for symbol in ('C41', 'phi41', 'C43', 'phi43', 'C45', 'phi41')]):
            array += (1 / 5 * alpha2 ** 2 * alpha *
                      (p['C41'] * xp.cos((phi - p['phi41'])) +
                       p['C43'] * xp.cos(3 * (phi - p['phi43'])) +
                       p['C45'] * xp.cos(5 * (phi - p['phi45']))))

        if any([p[symbol] != 0. for symbol in ('C50', 'C52', 'phi52', 'C54', 'phi54', 'C56', 'phi56')]):
            array += (1 / 6 * alpha2 ** 3 *
                      (p['C50'] +
                       p['C52'] * xp.cos(2 * (phi - p['phi52'])) +
                       p['C54'] * xp.cos(4 * (phi - p['phi54'])) +
                       p['C56'] * xp.cos(6 * (phi - p['phi56']))))

        array = 2 * xp.pi / self.wavelength * array
        return array

    def evaluate_aberrations(self, alpha: Union[float, np.ndarray], phi: Union[float, np.ndarray]) -> \
            Union[float, np.ndarray]:
        xp = get_array_module(alpha)
        complex_exponential = get_device_function(xp, 'complex_exponential')
        return complex_exponential(-self.evaluate_chi(alpha, phi))

    def evaluate(self, alpha: Union[float, np.ndarray], phi: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        array = self.evaluate_aberrations(alpha, phi)

        if (self.semiangle_cutoff < np.inf) or (self._aperture is not None):
            array *= self.evaluate_aperture(alpha, phi)

        if self.focal_spread > 0.:
            array *= self.evaluate_temporal_envelope(alpha)

        if self.angular_spread > 0.:
            array *= self.evaluate_spatial_envelope(alpha, phi)

        if self.gaussian_spread > 0.:
            array *= self.evaluate_gaussian_envelope(alpha)

        return array

    def _polar_coordinates(self, gpts=None, extent=None, sampling=None, xp=np):
        grid = Grid(gpts=gpts, extent=extent, sampling=sampling)

        gpts = grid.gpts
        sampling = grid.sampling

        kx, ky = spatial_frequencies(gpts, sampling)
        kx = kx.reshape((1, -1, 1))
        ky = ky.reshape((1, 1, -1))
        kx = xp.asarray(kx)
        ky = xp.asarray(ky)
        return polar_coordinates(xp.asarray(kx * self.wavelength), xp.asarray(ky * self.wavelength))

    def evaluate_on_grid(self, gpts=None, extent=None, sampling=None, xp=np):
        return self.evaluate(*self._polar_coordinates(gpts, extent, sampling, xp))

    def profiles(self, max_semiangle: float = None, phi: float = 0.):
        if max_semiangle is None:
            if self._semiangle_cutoff == np.inf:
                max_semiangle = 50
            else:
                max_semiangle = self._semiangle_cutoff * 1.6

        alpha = np.linspace(0, max_semiangle / 1000., 500)

        aberrations = self.evaluate_aberrations(alpha, phi)
        aperture = self.evaluate_aperture(alpha)
        temporal_envelope = self.evaluate_temporal_envelope(alpha)
        spatial_envelope = self.evaluate_spatial_envelope(alpha, phi)
        gaussian_envelope = self.evaluate_gaussian_envelope(alpha)
        envelope = aperture * temporal_envelope * spatial_envelope * gaussian_envelope

        calibration = Calibration(offset=0., sampling=(alpha[1] - alpha[0]) * 1000., units='mrad', name='alpha')

        profiles = {}
        profiles['ctf'] = Measurement(aberrations.imag * envelope, calibrations=[calibration], name='CTF')
        profiles['aperture'] = Measurement(aperture, calibrations=[calibration], name='Aperture')
        profiles['temporal_envelope'] = Measurement(temporal_envelope,
                                                    calibrations=[calibration],
                                                    name='Temporal')
        profiles['spatial_envelope'] = Measurement(spatial_envelope, calibrations=[calibration],
                                                   name='Spatial')
        profiles['gaussian_envelope'] = Measurement(gaussian_envelope, calibrations=[calibration],
                                    name='Gaussian')
        profiles['envelope'] = Measurement(envelope, calibrations=[calibration], name='Envelope')
        return profiles

    def apply(self, waves, interact=False, sliders=None, throttling=0.):
        if interact:
            from abtem.visualize.interactive import Canvas, MeasurementArtist2d
            from abtem.visualize.widgets import quick_sliders, throttle
            import ipywidgets as widgets

            image_waves = waves.copy()
            canvas = Canvas()
            artist = MeasurementArtist2d()
            canvas.artists = {'artist': artist}

            def update(*args):
                image_waves.array[:] = waves.apply_ctf(self).array
                artist.measurement = image_waves.intensity()[0]
                canvas.adjust_limits_to_artists()
                canvas.adjust_labels_to_artists()

            if throttling:
                update = throttle(throttling)(update)

            self.observe(update)
            update()

            if sliders:
                sliders = quick_sliders(self, **sliders)
                figure = widgets.HBox([canvas.figure, widgets.VBox(sliders)])
            else:
                figure = canvas.figure

            return image_waves, figure
        else:
            if sliders:
                raise RuntimeError()

            return waves.apply_ctf(self)

    def interact(self, max_semiangle: float = None, phi: float = 0., sliders=None, throttling=False):
        from abtem.visualize.interactive.utils import quick_sliders, throttle
        from abtem.visualize.interactive import Canvas, MeasurementArtist1d
        import ipywidgets as widgets

        canvas = Canvas(lock_scale=False)
        ctf_artist = MeasurementArtist1d()
        envelope_artist = MeasurementArtist1d()
        canvas.artists = {'ctf': ctf_artist, 'envelope': envelope_artist}
        canvas.y_scale.min = -1.1
        canvas.y_scale.max = 1.1

        def callback(*args):
            profiles = self.profiles(max_semiangle, phi)

            for name, artist in canvas.artists.items():
                artist.measurement = profiles[name]

        if throttling:
            callback = throttle(throttling)(callback)

        self.observe(callback)

        callback()
        canvas.adjust_limits_to_artists(adjust_y=False)
        canvas.adjust_labels_to_artists()

        if sliders:
            sliders = quick_sliders(self, **sliders)
            return widgets.HBox([canvas.figure, widgets.VBox(sliders)])
        else:
            return canvas.figure

    def show(self, max_semiangle: float = None, phi: float = 0, ax=None, **kwargs):
        """
        Show the contrast transfer function.

        Parameters
        ----------
        max_semiangle: float
            Maximum semiangle to display in the plot.
        ax: matplotlib Axes, optional
            If given, the plot will be added to this matplotlib axes.
        phi: float, optional
            The contrast transfer function will be plotted along this angle. Default is 0.
        n: int, optional
            Number of evaluation points to use in the plot. Default is 1000.
        title: str, optional
            The title of the plot. Default is 'None'.
        kwargs:
            Additional keyword arguments for the line plots.
        """
        import matplotlib.pyplot as plt

        if ax is None:
            ax = plt.subplot()

        for key, profile in self.profiles(max_semiangle, phi).items():
            if not np.all(profile.array == 1.):
                ax, lines = profile.show(legend=True, ax=ax, **kwargs)

        return ax

    def copy(self):
        parameters = self.parameters.copy()
        return self.__class__(semiangle_cutoff=self.semiangle_cutoff,
                              rolloff=self.rolloff,
                              focal_spread=self.focal_spread,
                              angular_spread=self.angular_spread,
                              gaussian_spread=self.gaussian_spread,
                              energy=self.energy,
                              parameters=parameters)


def scherzer_defocus(Cs, energy):
    """
    Calculate the Scherzer defocus.

    Parameters
    ----------
    Cs: float
        Spherical aberration [Å].
    energy: float
        Electron energy [eV].

    Returns
    -------
    float
        The Scherzer defocus.
    """

    return np.sign(Cs) * np.sqrt(3 / 2 * np.abs(Cs) * energy2wavelength(energy))


def point_resolution(Cs: float, energy: float):
    """
    Calculate the point resolution.

    Parameters
    ----------
    Cs: float
        Spherical aberration [Å].
    energy: float
        Electron energy [eV].

    Returns
    -------
    float
        The point resolution.
    """

    return (energy2wavelength(energy) ** 3 * np.abs(Cs) / 6) ** (1 / 4)


def polar2cartesian(polar):
    """
    Convert between polar and Cartesian aberration coefficients.

    Parameters
    ----------
    polar: dict
        Mapping from polar aberration symbols to their corresponding values.

    Returns
    -------
    dict
        Mapping from cartesian aberration symbols to their corresponding values.
    """

    polar = defaultdict(lambda: 0, polar)

    cartesian = dict()
    cartesian['C10'] = polar['C10']
    cartesian['C12a'] = - polar['C12'] * np.cos(2 * polar['phi12'])
    cartesian['C12b'] = polar['C12'] * np.sin(2 * polar['phi12'])
    cartesian['C21a'] = polar['C21'] * np.sin(polar['phi21'])
    cartesian['C21b'] = polar['C21'] * np.cos(polar['phi21'])
    cartesian['C23a'] = - polar['C23'] * np.sin(3 * polar['phi23'])
    cartesian['C23b'] = polar['C23'] * np.cos(3 * polar['phi23'])
    cartesian['C30'] = polar['C30']
    cartesian['C32a'] = - polar['C32'] * np.cos(2 * polar['phi32'])
    cartesian['C32b'] = polar['C32'] * np.cos(np.pi / 2 - 2 * polar['phi32'])
    cartesian['C34a'] = polar['C34'] * np.cos(-4 * polar['phi34'])
    K = np.sqrt(3 + np.sqrt(8.))
    cartesian['C34b'] = 1 / 4. * (1 + K ** 2) ** 2 / (K ** 3 - K) * polar['C34'] * np.cos(
        4 * np.arctan(1 / K) - 4 * polar['phi34'])

    return cartesian


def cartesian2polar(cartesian):
    """
    Convert between Cartesian and polar aberration coefficients.

    Parameters
    ----------
    cartesian: dict
        Mapping from Cartesian aberration symbols to their corresponding values.

    Returns
    -------
    dict
        Mapping from polar aberration symbols to their corresponding values.
    """

    cartesian = defaultdict(lambda: 0, cartesian)

    polar = dict()
    polar['C10'] = cartesian['C10']
    polar['C12'] = - np.sqrt(cartesian['C12a'] ** 2 + cartesian['C12b'] ** 2)
    polar['phi12'] = - np.arctan2(cartesian['C12b'], cartesian['C12a']) / 2.
    polar['C21'] = np.sqrt(cartesian['C21a'] ** 2 + cartesian['C21b'] ** 2)
    polar['phi21'] = np.arctan2(cartesian['C21a'], cartesian['C21b'])
    polar['C23'] = np.sqrt(cartesian['C23a'] ** 2 + cartesian['C23b'] ** 2)
    polar['phi23'] = -np.arctan2(cartesian['C23a'], cartesian['C23b']) / 3.
    polar['C30'] = cartesian['C30']
    polar['C32'] = -np.sqrt(cartesian['C32a'] ** 2 + cartesian['C32b'] ** 2)
    polar['phi32'] = -np.arctan2(cartesian['C32b'], cartesian['C32a']) / 2.
    polar['C34'] = np.sqrt(cartesian['C34a'] ** 2 + cartesian['C34b'] ** 2)
    polar['phi34'] = np.arctan2(cartesian['C34b'], cartesian['C34a']) / 4

    return polar



from abc import abstractmethod, ABC

import numpy as np
from bqplot.interacts import BrushSelector, PanZoom
from bqplot_image_gl.interacts import MouseInteraction
from traitlets import HasTraits, Int, default, Bool, List, Float, Instance, observe
from traittypes import Array

from abtem.visualize.interactive.artists import ScatterArtist, CircleArtist, LinesArtist
from abtem.visualize.interactive.utils import link


class Tool(ABC):

    @abstractmethod
    def activate(self, canvas):
        pass


class BoxZoomTool(Tool):

    def activate(self, canvas):
        brush_selector = BrushSelector()
        brush_selector.x_scale = canvas.x_scale
        brush_selector.y_scale = canvas.y_scale

        def box_zoom(change):
            selected_x = brush_selector.selected_x
            selected_y = brush_selector.selected_y

            if (selected_x is None) or (selected_y is None):
                return

            canvas.x_scale.min = min(selected_x[0], selected_x[1])
            canvas.x_scale.max = max(selected_x[0], selected_x[1])
            canvas.y_scale.min = min(selected_y[0], selected_y[1])
            canvas.y_scale.max = max(selected_y[0], selected_y[1])

            canvas._enforce_scale_lock()

            brush_selector.selected_x = None
            brush_selector.selected_y = None
            canvas.figure.interaction = None
            canvas.figure.interaction = brush_selector

        brush_selector.observe(box_zoom, 'brushing')
        canvas.figure.interaction = brush_selector

    def deactivate(self, canvas):
        canvas.figure.interaction = None


class PanZoomTool(Tool):

    def activate(self, canvas):
        panzoom = PanZoom()
        panzoom.scales = {'x': [canvas.x_scale], 'y': [canvas.y_scale]}
        canvas.figure.interaction = panzoom

    def deactivate(self, canvas):
        canvas.figure.interaction = None


class SelectPixelTool(HasTraits):
    __metaclass__ = Tool

    indices = List()

    # index_y = Int()

    def __init__(self, image_artist, **kwargs):
        self._image_artist = image_artist
        self._point_artist = ScatterArtist()

        super().__init__(**kwargs)

    @default('indices')
    def _default_indices(self):
        return [0, 0]

    def activate(self, canvas):
        self._point_artist._add_to_canvas(canvas)

        interaction = MouseInteraction(x_scale=canvas.figure.axes[0].scale,
                                       y_scale=canvas.figure.axes[1].scale)

        def on_mouse_msg(_, change, __):
            if change['event'] in ('dragmove', 'click'):
                position = np.array([change['domain']['x'], change['domain']['y']])
                indices = self._image_artist.position_to_index(position)
                indices[0] = max(0, min(indices[0], self._image_artist.image.shape[0] - 1))
                indices[1] = max(0, min(indices[1], self._image_artist.image.shape[1] - 1))

                self.indices = indices
                # self.index_y = indices[1]

                rounded_position = self._image_artist.indices_to_position((indices[0], indices[1]))
                self._point_artist.x = np.array([rounded_position[0]])
                self._point_artist.y = np.array([rounded_position[1]])

        canvas.figure.interaction = interaction
        interaction.events = ['click', 'dragmove']
        canvas.figure.interaction.on_msg(on_mouse_msg)

    def deactivate(self, canvas):
        artists = canvas.artists
        for name, artist in canvas.artists.items():
            if artist == self._point_artist:
                artists.pop(name)

        canvas.artists = artists
        canvas._observe_artists(None)
        canvas.figure.interaction = None


class BoxSelectPixelTool(HasTraits):
    __metaclass__ = Tool

    selected_x = List()
    selected_y = List()

    def __init__(self, image_artist, continuous_update=True, **kwargs):
        self._image_artist = image_artist
        self.continuous_update = continuous_update
        super().__init__(**kwargs)

    def activate(self, canvas):
        brush_selector = BrushSelector(color='red')
        brush_selector.x_scale = canvas.x_scale
        brush_selector.y_scale = canvas.y_scale

        def box_zoom(change):

            if not self.continuous_update:
                if (not change['name'] == 'brushing') and (change['new']):
                    return

            selected_x = brush_selector.selected_x
            selected_y = brush_selector.selected_y

            if (selected_x is None) or (selected_y is None):
                return

            corner1 = self._image_artist.position_to_index([selected_x[0], selected_y[0]])
            corner2 = self._image_artist.position_to_index([selected_x[1], selected_y[1]])

            for corner in (corner1, corner2):
                corner[0] = max(0, min(corner[0], self._image_artist.image.shape[0] - 1))
                corner[1] = max(0, min(corner[1], self._image_artist.image.shape[1] - 1))

            with self.hold_trait_notifications():
                self.selected_x = [min(corner1[0], corner2[0]), max(corner1[0], corner2[0])]
                self.selected_y = [min(corner1[1], corner2[1]), max(corner1[1], corner2[1])]

        brush_selector.observe(box_zoom)
        canvas.figure.interaction = brush_selector

    def deactivate(self, canvas):
        canvas.figure.interaction = None


class SelectPositionTool1d(HasTraits):
    position = Array()
    marker = Bool()

    def __init__(self, allow_drag=True, direction='x', **kwargs):
        self._point_artist = LinesArtist()
        self._allow_drag = allow_drag
        self._interaction = MouseInteraction()
        self._direction = direction
        super().__init__(**kwargs)

    @default('position')
    def _default_position(self):
        return 0.

    @observe('position')
    def _update_line(self, *args):
        if self.marker:
            if self._direction == 'y':
                self._point_artist.y = [self.position, self.position]
                self._point_artist.x = [-100, 100]
            else:
                self._point_artist.x = [self.position, self.position]
                self._point_artist.y = [-100, 100]

    def activate(self, canvas):
        if self.marker:
            canvas._tool_artists = [self._point_artist]

        self._interaction.x_scale = canvas.figure.axes[0].scale
        self._interaction.y_scale = canvas.figure.axes[1].scale

        def on_mouse_msg(_, change, __):
            if change['event'] in ('dragmove', 'click'):
                self.position = change['domain'][self._direction]

        canvas.figure.interaction = self._interaction

        events = ['click']
        if self._allow_drag:
            events += ['dragmove']

        self._interaction.events = events
        self._interaction.on_msg(on_mouse_msg)

    def deactivate(self, canvas):
        canvas._tool_artists = []
        canvas.figure.interaction = None


class DragPointTool(HasTraits):
    artist = Instance(ScatterArtist)

    def activate(self, canvas):
        self.artist._mark.enable_move = True

    def deactivate(self, canvas):
        self.artist._mark.enable_move = False


class SelectRadiusTool(HasTraits):
    radius = Float(1.)
    always_visibile = Bool()

    def __init__(self, **kwargs):
        self._circle_artist = CircleArtist()
        self._select_position_tool = SelectPositionTool(marker=False)

        super().__init__(**kwargs)
        link((self, 'radius'), (self._circle_artist, 'radius'))

        def callback(*args):
            radius = min(np.linalg.norm(self._circle_artist.center - self._select_position_tool.position), 1)
            self.radius = radius

        self._callback = callback

    def activate(self, canvas):
        self._select_position_tool.activate(canvas)
        self._select_position_tool.observe(self._callback)
        canvas._tool_artists = [self._circle_artist]

    def deactivate(self, canvas):
        self._select_position_tool.unobserve(self._callback)
        # canvas._tool_artists = []


class SelectAnnularRadiiTool(HasTraits):
    center = Array()
    inner_radius = Float(1.)
    outer_radius = Float(2.)
    min_width = Float(0.)
    max_radius = Float(2.1)
    lock_radii = Bool()

    def __init__(self, **kwargs):
        self._outer_circle_artist = CircleArtist()
        self._inner_circle_artist = CircleArtist()
        self._mouse_interaction = MouseInteraction()
        self._width = None
        self._adjusting = 'inner'

        self._observe_radii()

        super().__init__(**kwargs)

        link((self, 'inner_radius'), (self._inner_circle_artist, 'radius'))
        link((self, 'outer_radius'), (self._outer_circle_artist, 'radius'))
        link((self, 'center'), (self._inner_circle_artist, 'center'))
        link((self, 'center'), (self._outer_circle_artist, 'center'))

        def callback(_, change, __):
            position = np.array([change['domain']['x'], change['domain']['y']])
            distance = np.linalg.norm(self.center - position)
            if change['event'] == 'dragstart':
                inner_distance = np.abs(distance - self._inner_circle_artist.radius)
                outer_distance = np.abs(distance - self._outer_circle_artist.radius)
                if inner_distance < outer_distance:
                    self._adjusting = 'inner'
                else:
                    self._adjusting = 'outer'

            if change['event'] in ('dragstart', 'dragmove'):
                if self._adjusting == 'inner':
                    with self.hold_trait_notifications():
                        new_inner_radius = min(distance, self.outer_radius - self.min_width)
                        if self.lock_radii:
                            self.inner_radius = min(new_inner_radius, self.max_radius - self._width)
                            self.outer_radius = self.inner_radius + self._width
                        else:
                            self.inner_radius = new_inner_radius

                else:
                    with self.hold_trait_notifications():
                        new_outer_radius = max(min(distance, self.max_radius), self.inner_radius + self.min_width)
                        if self.lock_radii:
                            self.outer_radius = max(new_outer_radius, self._width)
                            self.inner_radius = self.outer_radius - self._width

                        else:
                            self.outer_radius = new_outer_radius

        self._mouse_interaction.events = ['dragstart', 'dragmove']
        self._callback = callback

    @default('center')
    def _default_center(self):
        return np.zeros((2,))

    @observe('inner_radius', 'outer_radius')
    def _observe_radii(self, *args):
        self._width = self.outer_radius - self.inner_radius

    def activate(self, canvas):
        self._mouse_interaction.x_scale = canvas.figure.axes[0].scale
        self._mouse_interaction.y_scale = canvas.figure.axes[1].scale
        self._mouse_interaction.on_msg(self._callback)
        canvas.figure.interaction = self._mouse_interaction
        canvas._tool_artists = [self._inner_circle_artist, self._outer_circle_artist]

    def deactivate(self, canvas):
        self._mouse_interaction.unobserve(self._callback)
        canvas._tool_artists = []



from ase.data import chemical_symbols

from abtem.utils import _set_path

azimuthal_number = {'s': 0, 'p': 1, 'd': 2, 'f': 3, 'g': 4, 'h': 5, 'i': 6}
azimuthal_letter = {value: key for key, value in azimuthal_number.items()}


def load_electronic_configurations():
    configurations = {}
    with open(_set_path('electron_configurations.txt')) as f:
        for i, line in enumerate(f):
            line = line.strip()
            prefix_start = line.find('[')
            prefix_end = line.find(']')
            if prefix_start > -1:
                line = configurations[line[prefix_start + 1:prefix_end]] + line[prefix_end + 1:]

            configurations[chemical_symbols[i + 1]] = line
    return configurations


def config_str_to_config_tuples(config_str):
    config_tuples = []
    for subshell_string in config_str.split(' '):
        config_tuples.append((int(subshell_string[0]), azimuthal_number[subshell_string[1]], int(subshell_string[2])))
    return config_tuples


def config_tuples_to_config_str(config_tuples):
    config_str = []
    for n, ell, occ in config_tuples:
        config_str.append(str(n) + azimuthal_letter[ell] + str(occ))
    return ' '.join(config_str)


def remove_electron_from_config_str(config_str, n, ell):
    config_tuples = []
    for shell in config_str_to_config_tuples(config_str):
        if shell[:2] == (n, ell):
            config_tuples.append(shell[:2] + (shell[2] - 1,))
        else:
            config_tuples.append(shell)
    return config_tuples_to_config_str(config_tuples)


def check_valid_quantum_number(Z, n, ell):
    symbol = chemical_symbols[Z]
    config_tuple = config_str_to_config_tuples(load_electronic_configurations()[symbol])

    if not any([shell[:2] == (n, ell) for shell in config_tuple]):
        raise RuntimeError(f'Quantum numbers (n, ell) = ({n}, {ell}) not valid for element {symbol}')
